"""
volmdlr utils for calculating 3D to surface parametric domain operation.

"""
import bisect
import math

import volmdlr
import volmdlr.edges as vme
from volmdlr import curves


def find_sign_changes(list_of_values):
    """
    Finds the position of sign changes in a list.

    :param list_of_values: The list to search for sign changes.
    :type list_of_values: list
    :returns: A list of indices where the sign changes occur.
    :rtype: list
    """
    sign_changes = []
    for i in range(1, len(list_of_values)):
        if list_of_values[i] * list_of_values[i - 1] < 0:
            sign_changes.append(i)
    return sign_changes


def angle_discontinuity(angle_list):
    """
    Returns True if there is some angle discontinuity in the angle_list.

    This discontinuity can occur when we transform some 3D primitive into the parametric domain of periodical surfaces.

    :param angle_list: List with angle axis values of the primitive in parametric domain.
    :type angle_list: List[float]
    :return: Returns True if there is discontinuity, False otherwise.
    :rtype: bool
    """
    indexes_sign_changes = find_sign_changes(angle_list)
    discontinuity = False
    indexes_angle_discontinuity = []
    if indexes_sign_changes:
        for index in indexes_sign_changes:
            sign = round(angle_list[index - 1] / abs(angle_list[index - 1]), 2)
            delta = max(abs(angle_list[index] + sign * volmdlr.TWO_PI - angle_list[index - 1]), 1e-4)
            if math.isclose(abs(angle_list[index]), math.pi, abs_tol=1.1 * delta) and \
                    not math.isclose(abs(angle_list[index]), 0, abs_tol=1.1 * delta):
                indexes_angle_discontinuity.append(index)
                discontinuity = True
    return discontinuity, indexes_angle_discontinuity


def is_undefined_brep_primitive(primitive, periodicity):
    """
    2D primitives on parametric surface domain can be in wrong bound side due to periodic behavior of some surfaces.

    A B-Rep primitive can be undefined when it's not directly provided, but it's found by some 3D to 2D operation.
    This can result in a primitive that is entirely contained on the periodical boundary, and we can only know its
    right position when it's placed on the boundary representation, by analyzing the continuity of the 2D contour.

    :param primitive: primitive to perform verification.
    :type primitive: vme.Edge
    :param periodicity: list with periodicity in x and y direction
    :type periodicity: list
    """
    start = primitive.start
    end = primitive.end

    if periodicity[0] and not periodicity[1]:
        i = 0
    elif not periodicity[0] and periodicity[1]:
        i = 1
    else:
        return False
    #     if ((2 * start.x) % periodicity[0]) == 0 and ((2 * start.y) % periodicity[1]) == 0:
    #         return True
    #     return False
    if ((2 * start[i]) % periodicity[i]) == 0 and end[i] == start[i]:
        return True
    return False


def find_index_defined_brep_primitive_on_periodical_surface(primitives2d, periodicity):
    """
    Search for a primitive on the boundary representation that can be used as reference for repairing the contour.
    """
    pos = 0
    for i, primitive in enumerate(primitives2d):
        if not is_undefined_brep_primitive(primitive, periodicity):
            return i
    return pos


def repair_singularity(primitive, last_primitive):
    """
    Repairs the Contour2D of SphericalSurface3D and ConicalSurface3D parametric face representations.

    Used when transforming from spatial to parametric coordinates when the surface contains a singularity
    """
    v1 = primitive.unit_direction_vector()
    v2 = last_primitive.unit_direction_vector()
    dot = v1.dot(volmdlr.X2D)
    cross = v1.cross(v2)
    new_primitives = []
    if cross == 0 and dot == 0:
        if primitive.start.x == math.pi:
            primitive = primitive.translation(volmdlr.Vector2D(-2 * math.pi, 0))
            new = vme.LineSegment2D(last_primitive.end, primitive.start)
        elif primitive.start.x == -math.pi:
            primitive = primitive.translation(volmdlr.Vector2D(2 * math.pi, 0))
            new = vme.LineSegment2D(last_primitive.end, primitive.start)
        else:
            new = vme.LineSegment2D(last_primitive.end, primitive.start)

        new_primitives.append(new)
        new_primitives.append(primitive)
    else:
        delta = last_primitive.end - primitive.start
        new_primitives.append(primitive.translation(delta))
    return new_primitives


def repair_start_end_angle_periodicity(angle, ref_angle):
    """
    Repairs start and end angles in parametric coordinates.

    Uses ref_angle (angle just after start angle, if repairing start angle or angle just before end if repairing end
        angle).

    :param angle: Angle to repair.
    :type angle: float
    :param ref_angle: Angle of reference.
    :return: Returns the repaired angle.
    :rtype: float
    """
    # Verify if theta1 or theta2 point should be -pi because atan2() -> ]-pi, pi]
    if math.isclose(angle, math.pi, abs_tol=1e-3) and ref_angle < 0:
        angle = -math.pi
    elif math.isclose(angle, -math.pi, abs_tol=1e-3) and ref_angle > 0:
        angle = math.pi
    elif math.isclose(angle, 0.5 * math.pi, abs_tol=1e-3) and ref_angle < 0:
        angle = -0.5 * math.pi
    elif math.isclose(angle, -0.5 * math.pi, abs_tol=1e-3) and ref_angle > 0:
        angle = 0.5 * math.pi
    return angle


def repair_arc3d_angle_continuity(angle_start, angle_end, periodicity):
    """
    Repairs Arc3D continuity after conversion of points to parametric 2D space.
    """
    if angle_start < angle_end:
        angle_end -= periodicity
    else:
        angle_end += periodicity

    return angle_start, angle_end


def arc3d_to_cylindrical_coordinates_verification(start_end, start_end_theta_state, reference_points,
                                                  discontinuity):
    """
    Verifies theta from start and end of an Arc3D after transformation from spatial to parametric coordinates.
    """
    start, end = start_end
    theta1, z1 = start
    theta2, z2 = end
    undefined_start_theta, undefined_end_theta = start_end_theta_state
    theta3, theta4 = reference_points

    if undefined_start_theta or abs(theta1) == 0.5 * math.pi:
        theta1 = repair_start_end_angle_periodicity(theta1, theta3)
    if undefined_end_theta or abs(theta2) == 0.5 * math.pi:
        theta2 = repair_start_end_angle_periodicity(theta2, theta4)

    if discontinuity:
        theta1, theta2 = repair_arc3d_angle_continuity(theta1, theta2, volmdlr.TWO_PI)

    start = volmdlr.Point2D(theta1, z1)
    end = volmdlr.Point2D(theta2, z2)
    return [start, end]


def fullarc_to_cylindrical_coordinates_verification(start, end, normal_dot_product):
    """
    Verifies theta from start and end of a full arc after transformation from spatial to parametric coordinates.
    """
    theta1, z1 = start
    _, z2 = end
    if normal_dot_product > 0:
        end = volmdlr.Point2D(theta1 + volmdlr.TWO_PI, z2)
    elif normal_dot_product < 0 and math.isclose(theta1, -math.pi, abs_tol=1e-6):
        start = volmdlr.Point2D(math.pi, z1)
    elif normal_dot_product < 0:
        end = volmdlr.Point2D(theta1 - volmdlr.TWO_PI, z2)
    return [start, end]


def toroidal_repair_start_end_angle_periodicity(start, end, start_end_angles_state,
                                                point_after_start, point_before_end):
    """
    Verifies theta and phi from start and end of an arc 3D after transformation from spatial to parametric coordinates.
    """
    theta1, phi1 = start
    theta2, phi2 = end
    undefined_start_theta, undefined_end_theta, undefined_start_phi, undefined_end_phi = start_end_angles_state
    # Verify if theta1 or theta2 point should be -pi or pi because atan2() -> ]-pi, pi]
    if undefined_start_theta or abs(theta1) == 0.5 * math.pi:
        theta1 = repair_start_end_angle_periodicity(theta1, point_after_start.x)
    if undefined_end_theta or abs(theta2) == 0.5 * math.pi:
        theta2 = repair_start_end_angle_periodicity(theta2, point_before_end.x)

    # Verify if phi1 or phi2 point should be -pi or pi because phi -> ]-pi, pi]
    if undefined_start_phi or abs(phi1) == 0.5 * math.pi:
        phi1 = repair_start_end_angle_periodicity(phi1, point_after_start.y)
    if undefined_end_phi or abs(phi2) == 0.5 * math.pi:
        phi2 = repair_start_end_angle_periodicity(phi2, point_before_end.y)
    return volmdlr.Point2D(theta1, phi1), volmdlr.Point2D(theta2, phi2)


def spherical_repair_start_end_angle_periodicity(start, end, point_after_start, point_before_end):
    """
    Verifies theta and phi from start and end of an arc 3D after transformation from spatial to parametric coordinates.
    """
    theta1, phi1 = start
    theta2, phi2 = end
    theta3, _ = point_after_start
    theta4, _ = point_before_end
    # Verify if theta1 or theta2 point should be -pi or pi because atan2() -> ]-pi, pi]
    if abs(theta1) == math.pi or abs(theta1) == 0.5 * math.pi:
        theta1 = repair_start_end_angle_periodicity(theta1, theta3)
    if abs(theta2) == math.pi or abs(theta2) == 0.5 * math.pi:
        theta2 = repair_start_end_angle_periodicity(theta2, theta4)

    return volmdlr.Point2D(theta1, phi1), volmdlr.Point2D(theta2, phi2)


def arc3d_to_toroidal_coordinates_verification(start_end, start_end_angles_state, reference_points,
                                               discontinuity):
    """
    Verifies theta and phi from start and end of an arc 3D after transformation from spatial to parametric coordinates.
    """
    start, end = start_end
    point_after_start, point_before_end = reference_points
    start, end = toroidal_repair_start_end_angle_periodicity(start, end, start_end_angles_state, point_after_start,
                                                             point_before_end)
    theta1, phi1 = start
    theta2, phi2 = end
    theta_discontinuity, phi_discontinuity = discontinuity
    if theta_discontinuity:
        theta1, theta2 = repair_arc3d_angle_continuity(theta1, theta2, volmdlr.TWO_PI)

    if phi_discontinuity:
        phi1, phi2 = repair_arc3d_angle_continuity(phi1, phi2, volmdlr.TWO_PI)

    return volmdlr.Point2D(theta1, phi1), volmdlr.Point2D(theta2, phi2)


def arc3d_to_spherical_coordinates_verification(start_end, reference_points, discontinuity):
    """
    Verifies theta and phi from start and end of an arc 3D after transformation from spatial to parametric coordinates.
    """
    start, end = start_end
    point_after_start = reference_points[0]
    point_before_end = reference_points[1]
    start, end = spherical_repair_start_end_angle_periodicity(start, end, point_after_start, point_before_end)
    theta1, phi1 = start
    theta2, phi2 = end
    if discontinuity and math.isclose(phi1, phi2, abs_tol=1e-4):
        theta1, theta2 = repair_arc3d_angle_continuity(theta1, theta2, volmdlr.TWO_PI)

    return volmdlr.Point2D(theta1, phi1), volmdlr.Point2D(theta2, phi2)


def array_range_search(x, xmin, xmax):
    """
    Find the indices of the elements in the sorted list `x` that fall within the specified range.

    This function use bisect python builtin module, which uses binary search and has a time complexity of O(log(n)).
    Where n is the array length.

    :param x: A sorted list of values.
    :type x: list
    :param xmin: The minimum value in the range.
    :type xmin: float
    :param xmax: The maximum value in the range.
    :type xmax: float
    :return: A python range from the first to the last elements in `x`.
    :rtype: range
    :Example:

    >>> x = [1, 2, 3, 4, 5]
    >>> array_range_search(x, 2, 4)
    range(1, 3)
    """

    left = bisect.bisect_left(x, xmin)
    right = bisect.bisect_right(x, xmax)

    return range(left, right)


def contour2d_healing(contour2d):
    """
    Heals topologies incoherencies on the boundary representation.
    """
    contour2d = contour2d_healing_self_intersection(contour2d)
    # contour2d = contour2d_healing_close_gaps(contour2d, contour3d)
    return contour2d


def contour2d_healing_close_gaps(contour2d, contour3d):
    """
    Heals topologies incoherencies on the boundary representation.
    """
    new_primitives = []
    for prim1_3d, prim2_3d, prim1, prim2 in zip(contour3d.primitives, contour3d.primitives[1:] +
                                                [contour3d.primitives[0]],
                                                contour2d.primitives, contour2d.primitives[1:] +
                                                                      [contour2d.primitives[0]]):
        if prim1 and prim2:
            if not prim1_3d.end.is_close(prim2_3d.start) and not prim1.end.is_close(prim2.start):
                new_primitives.append(vme.LineSegment2D(prim1.end, prim2.start))

            new_primitives.append(prim2)
    contour2d.primitives = new_primitives

    return contour2d


def contour2d_healing_self_intersection(contour2d):
    """
    Heals topologies incoherencies on the boundary representation.
    """
    primitives = contour2d.primitives
    for i, (prim1, prim2) in enumerate(
            zip(contour2d.primitives, contour2d.primitives[1:] + [contour2d.primitives[0]])):
        if not prim1.end.is_close(prim2.start):
            # check intersection
            intersections = prim1.intersections(prim2)
            if intersections:
                if len(intersections) > 1:
                    intersections = prim1.sort_points_along_curve(intersections)
                split_point = intersections[0]
                if prim1.is_point_edge_extremity(split_point):
                    new_prim1 = prim1
                else:
                    new_prim1 = prim1.split(split_point)[0]
                if prim2.is_point_edge_extremity(split_point):
                    new_prim2 = prim2
                else:
                    new_prim2 = prim2.split(split_point)[1]
                primitives[i] = new_prim1
                primitives[(i + 1) % len(contour2d.primitives)] = new_prim2
    contour2d.primitives = primitives
    return contour2d


def find_parametric_point_at_singularity(edge, reference_point, singularity_line):
    """Uses tangent line to find real theta angle of the singularity point on parametric domain."""
    abscissa_before_singularity = edge.abscissa(reference_point)
    direction_vector = edge.direction_vector(abscissa_before_singularity)
    direction_line = curves.Line2D(reference_point, reference_point + direction_vector)
    return direction_line.line_intersections(singularity_line)[0]



"""Open rounded line segments model."""
import random

import volmdlr
from volmdlr import primitives3d

random.seed(2)

p1 = volmdlr.Point3D(0, 0, 0)
p2 = volmdlr.Point3D(-0.150, 0, 0)
p3 = volmdlr.Point3D(-0.150, 0.215, 0)
p4 = volmdlr.Point3D(-0.150, 0.215, -0.058)
p5 = volmdlr.Point3D(-0.220, 0.186, -0.042)

points = [p1, p2, p3, p4, p5]
radius = {1: 0.015, 2: 0.020, 3: 0.03}

current_point = p5

for i in range(6):
    current_point += volmdlr.Point3D.random(-0.1, 0.3, -0.1, 0.3, -0.1, 0.3)
    points.append(current_point)
    radius[4 + i] = 0.01 + 0.03 * random.random()


open_rounded_line_segements = primitives3d.OpenRoundedLineSegments3D(points, radius, adapt_radius=True, name='wire')



"""
volmdlr utils for translating volmdlr primitives into step file entity.
"""


def product_writer(current_id, product_name):
    """
    Helper function to write the step definitions of a product.
    """
    step_content = ''
    product_definition_context_id = current_id + 1
    step_content += (f"#{product_definition_context_id} = "
                     + "PRODUCT_DEFINITION_CONTEXT('part definition',#2,'design');\n")

    product_context_id = product_definition_context_id + 1
    step_content += f"#{product_context_id} = PRODUCT_CONTEXT('',#2,'mechanical');\n"
    product_id = product_context_id + 1
    step_content += f"#{product_id} = PRODUCT('{product_name}'," \
                    f"'{product_name}','',(#{product_context_id}));\n"
    product_definition_formation_id = product_id + 1
    step_content += f"#{product_definition_formation_id} = " \
                    f"PRODUCT_DEFINITION_FORMATION('','',#{product_id});\n"
    product_definition_id = product_definition_formation_id + 1
    step_content += f"#{product_definition_id} = PRODUCT_DEFINITION('design'," \
                    f"'',#{product_definition_formation_id},#{product_definition_context_id});\n"
    product_definition_shape_id = product_definition_id + 1
    step_content += f"#{product_definition_shape_id} = PRODUCT_DEFINITION_SHAPE(''," \
                    f"'',#{product_definition_id});\n"
    shape_definition_repr_id = product_definition_shape_id + 1
    shape_representation_id = shape_definition_repr_id + 1
    step_content += f"#{shape_definition_repr_id} = SHAPE_DEFINITION_REPRESENTATION(" \
                    f"#{product_definition_shape_id},#{shape_representation_id});\n"

    return step_content, shape_definition_repr_id


def geometric_context_writer(current_id, uncertainty: float = 1e-4):
    """
    Helper function to write the step definitions of a product.
    """
    step_content = ''
    legth_unit_id = current_id + 1
    plane_angle_unit_id = legth_unit_id + 1
    solid_angle_unit = plane_angle_unit_id + 1
    uncertainty_id = solid_angle_unit + 1
    geometric_representation_context_id = uncertainty_id + 1

    step_content += f"#{legth_unit_id} = ( LENGTH_UNIT() NAMED_UNIT(*) SI_UNIT(.MILLI.,.METRE.) );\n"
    step_content += f"#{plane_angle_unit_id} = ( NAMED_UNIT(*) PLANE_ANGLE_UNIT() SI_UNIT($,.RADIAN.) );\n"
    step_content += f"#{solid_angle_unit} = ( NAMED_UNIT(*) SI_UNIT($,.STERADIAN.) SOLID_ANGLE_UNIT() );\n"
    step_content += f"#{uncertainty_id} = UNCERTAINTY_MEASURE_WITH_UNIT(LENGTH_MEASURE({uncertainty})," \
                    f"#{legth_unit_id},'distance_accuracy_value','confusion accuracy');\n"
    step_content += f"#{geometric_representation_context_id} = ( GEOMETRIC_REPRESENTATION_CONTEXT(3) " \
                    f"GLOBAL_UNCERTAINTY_ASSIGNED_CONTEXT((#{uncertainty_id})) " \
                    f"GLOBAL_UNIT_ASSIGNED_CONTEXT((#{legth_unit_id},#{plane_angle_unit_id}," \
                    f"#{solid_angle_unit})) " \
                    f"REPRESENTATION_CONTEXT('Context #1','3D Context with UNIT and UNCERTAINTY') );\n"

    return step_content, geometric_representation_context_id


def assembly_definition_writer(current_id, assembly_data, component_data, assembly_frame_id, component_frame_id):
    """Returns component assembly data."""
    shape_representation, assembly_product_definition = assembly_data
    component_shape_representation, component_product_definition = component_data
    step_content = ''
    context_dependent_id = current_id + 1
    repr_relationship_with_transfo_id = context_dependent_id + 1
    item_defined_transfo_id = repr_relationship_with_transfo_id + 1
    product_definition_shape_id = item_defined_transfo_id + 1
    next_assembly_usage_occurrence_id = product_definition_shape_id + 1

    step_content += f"#{context_dependent_id} = CONTEXT_DEPENDENT_SHAPE_REPRESENTATION(" \
                    f"#{repr_relationship_with_transfo_id},#{product_definition_shape_id});\n"
    step_content += f"#{repr_relationship_with_transfo_id} = ( REPRESENTATION_RELATIONSHIP('',''," \
                    f"#{component_shape_representation},#{shape_representation}) " \
                    f"REPRESENTATION_RELATIONSHIP_WITH_TRANSFORMATION(#{item_defined_transfo_id}) " \
                    f"SHAPE_REPRESENTATION_RELATIONSHIP() );\n"
    step_content += f"#{item_defined_transfo_id} = ITEM_DEFINED_TRANSFORMATION('','',#{assembly_frame_id}," \
                    f"#{component_frame_id});\n"
    step_content += f"#{product_definition_shape_id} = PRODUCT_DEFINITION_SHAPE('Placement'," \
                    f"'Placement of an item',#{next_assembly_usage_occurrence_id});\n"
    step_content += f"#{next_assembly_usage_occurrence_id} = NEXT_ASSEMBLY_USAGE_OCCURRENCE(" \
                    f"'','','',#{assembly_product_definition},#{component_product_definition},$);\n"
    return step_content, next_assembly_usage_occurrence_id


def step_ids_to_str(ids):
    """
    Returns a string with a '#' in front of each ID and a comma separating each-one.

    :param ids: A list of step primitives IDs
    :type ids: List[int]
    :return: A string containing all the IDs
    :rtype: str
    """
    return ','.join([f"#{i}" for i in ids])


STEP_HEADER = '''ISO-10303-21;
HEADER;
FILE_DESCRIPTION(('{name}'),'2;1');
FILE_NAME('{filename}','{timestamp}',('Author'),(''),'Volmdlr v{version}','','Unknown');
FILE_SCHEMA(('AUTOMOTIVE_DESIGN {{ 1 0 10303 214 1 1 1 1 }}'));
ENDSEC;
DATA;
#1 = APPLICATION_PROTOCOL_DEFINITION('international standard','automotive_design',2000,#2);
#2 = APPLICATION_CONTEXT('core data for automotive mechanical design processes');
'''

STEP_FOOTER = '''ENDSEC;
END-ISO-10303-21;
'''



"""
volmdlr utils for calculating curves intersections.

"""
import math

import volmdlr
from volmdlr.utils.common_operations import get_abscissa_discretization, get_plane_equation_coefficients


def circle_3d_line_intersections(circle_3d, line, abs_tol: float = 1e-6):
    """
    Calculates the intersections between a Circle3D and a Line3D.

    :param circle_3d: Circle3D or Arc3D
    :param line: Line3D to verify intersections
    :param abs_tol: Tolerance.
    :return: list of points intersecting Circle
    """
    intersections = []
    if not math.isclose(abs(circle_3d.frame.w.dot(volmdlr.Z3D)), 1, abs_tol=abs_tol):
        frame_mapped_circle = circle_3d.frame_mapping(circle_3d.frame, 'new')
        frame_mapped_line = line.frame_mapping(circle_3d.frame, 'new')
        circle_linseg_intersections = circle_3d_line_intersections(frame_mapped_circle, frame_mapped_line)
        for inter in circle_linseg_intersections:
            intersections.append(circle_3d.frame.local_to_global_coordinates(inter))
        return intersections
    distance_center_lineseg = line.point_distance(circle_3d.frame.origin)
    if distance_center_lineseg > circle_3d.radius:
        return []
    direction_vector = line.direction_vector()
    if math.isclose(line.point1.z, line.point2.z, abs_tol=1e-6) and \
            math.isclose(line.point2.z, circle_3d.frame.origin.z, abs_tol=abs_tol):
        if line.point1.is_close(circle_3d.center):
            point1 = line.point2
            vec = line.point1 - line.point2
        else:
            point1 = line.point1
            vec = line.point2 - line.point1
        quadratic_equation_a = vec.dot(vec)
        quadratic_equation_b = 2 * vec.dot(point1 - circle_3d.center)
        quadratic_equation_c = point1.dot(point1) + circle_3d.center.dot(circle_3d.center) - 2 * point1.dot(
            circle_3d.center) - circle_3d.radius ** 2

        delta = quadratic_equation_b ** 2 - 4 * quadratic_equation_a * quadratic_equation_c
        if delta < 0:  # No real solutions, no intersection
            return []
        if math.isclose(delta, 0, abs_tol=1e-7):  # One real solution, tangent intersection
            t_param = -quadratic_equation_b / (2 * quadratic_equation_a)
            return [point1 + t_param * vec]
        sqrt_delta = math.sqrt(delta)
        t_param = (-quadratic_equation_b + sqrt_delta) / (2 * quadratic_equation_a)
        s_param = (-quadratic_equation_b - sqrt_delta) / (2 * quadratic_equation_a)
        return [point1 + t_param * vec, point1 + s_param * vec]
    z_constant = circle_3d.frame.origin.z
    constant = (z_constant - line.point1.z) / direction_vector.z
    x_coordinate = constant * direction_vector.x + line.point1.x
    y_coordinate = constant * direction_vector.y + line.point1.y
    if math.isclose((x_coordinate - circle_3d.frame.origin.x) ** 2 + (y_coordinate - circle_3d.frame.origin.y) ** 2,
                    circle_3d.radius ** 2, abs_tol=1e-6):
        intersections = [volmdlr.Point3D(x_coordinate, y_coordinate, z_constant)]
    return intersections


def conic3d_line_intersections(conic3d, line3d, abs_tol: float = 1e-6):
    """
    Calculates the intersections between an Ellipse3D and a Line3D.

    :param conic3d: The Hyperbola 3D.
    :param line3d: The Line 3D.
    :param abs_tol: Tolerance.
    :return: list of points intersecting the Hyperbola 3D.
    """
    intersections = []
    if not math.isclose(abs(conic3d.frame.w.dot(volmdlr.Z3D)), 1, abs_tol=abs_tol):
        frame_mapped_conic3d = conic3d.frame_mapping(conic3d.frame, 'new')
        frame_mapped_line = line3d.frame_mapping(conic3d.frame, 'new')
        circle_linseg_intersections = conic3d_line_intersections(frame_mapped_conic3d, frame_mapped_line)
        for inter in circle_linseg_intersections:
            intersections.append(conic3d.frame.local_to_global_coordinates(inter))
        return intersections

    if line3d.point1.z == line3d.point2.z == conic3d.frame.origin.z:
        conic2d = conic3d.self_2d
        line2d = line3d.to_2d(conic3d.frame.origin, conic3d.frame.u, conic3d.frame.v)
        intersections_2d = conic2d.line_intersections(line2d)
        for intersection in intersections_2d:
            intersections.append(volmdlr.Point3D(intersection[0], intersection[1], conic3d.frame.origin.z))
        return intersections
    plane_lineseg_intersections = get_plane_line_intersections(conic3d.frame, line3d)
    if conic3d.point_belongs(plane_lineseg_intersections[0]):
        return plane_lineseg_intersections
    return []


def ellipse2d_line_intersections(ellipse2d, line2d):
    """
    Calculates the intersections between a line and an ellipse.

    :param ellipse2d: Ellipse to calculate intersections
    :param line2d: line to calculate intersections
    :return: list of points intersections, if there are any
    """
    if line2d.point_distance(ellipse2d.center) > ellipse2d.major_axis + 1e-6:
        return []
    theta = volmdlr.geometry.clockwise_angle(ellipse2d.major_dir, volmdlr.X2D)
    if not math.isclose(theta, 0.0, abs_tol=1e-6) and not math.isclose(theta, 2 * math.pi, abs_tol=1e-6):
        frame = volmdlr.Frame2D(ellipse2d.center, ellipse2d.major_dir, ellipse2d.minor_dir)
        frame_mapped_ellipse = ellipse2d.frame_mapping(frame, 'new')
        line_inters = frame_mapped_ellipse.line_intersections(line2d.frame_mapping(frame, 'new'))
        line_intersections = [frame.local_to_global_coordinates(point) for point in line_inters]
        return line_intersections

    if line2d.point2.x == line2d.point1.x:
        x1 = line2d.point1.x
        x2 = x1
        y1 = ellipse2d.minor_axis * math.sqrt((1 - x1 ** 2 / ellipse2d.major_axis ** 2))
        y2 = -y1
    else:
        m = (line2d.point2.y - line2d.point1.y) / (line2d.point2.x - line2d.point1.x)
        c = - m * (line2d.point1.x + ellipse2d.center.x) + line2d.point1.y + ellipse2d.center.y
        if ellipse2d.major_axis ** 2 * m ** 2 + ellipse2d.minor_axis ** 2 > c ** 2:
            x1 = - (2 * (ellipse2d.major_axis ** 2) * m * c + math.sqrt(
                (2 * (ellipse2d.major_axis ** 2) * m * c) ** 2 - 4 * (
                        ellipse2d.major_axis ** 2 * m ** 2 + ellipse2d.minor_axis ** 2) *
                ellipse2d.major_axis ** 2 * (c ** 2 - ellipse2d.minor_axis ** 2))) / (
                         2 * (ellipse2d.major_axis ** 2 * (m ** 2) +
                              ellipse2d.minor_axis ** 2))

            x2 = - (2 * (ellipse2d.major_axis ** 2) * m * c - math.sqrt(
                (2 * (ellipse2d.major_axis ** 2) * m * c) ** 2 - 4 * (
                        ellipse2d.major_axis ** 2 * m ** 2 + ellipse2d.minor_axis ** 2) *
                ellipse2d.major_axis ** 2 * (c ** 2 - ellipse2d.minor_axis ** 2))) / (
                         2 * (ellipse2d.major_axis ** 2 * (m ** 2) +
                              ellipse2d.minor_axis ** 2))
            y1 = m * x1 + c
            y2 = m * x2 + c
            point1 = volmdlr.Point2D(x1, y1)
            point2 = volmdlr.Point2D(x2, y2)
            if point1 == point2:
                return [point1]
            return [point1, point2]
    return []


def get_circle_intersections(circle1, circle2):
    """
    Calculates the intersections between two circle 2d.

    :param circle1: circle 1 verify intersection with bspline
    :param circle2: circle 2 to search for intersections.
    :return: a list with all intersections between the two circles.
    """
    if circle1.center.is_close(circle2.center):
        return []
    x0, y0 = circle1.center
    x1, y1 = circle2.center

    d_param = math.sqrt((x1 - x0) ** 2 + (y1 - y0) ** 2)

    # non-intersecting
    if d_param > circle1.radius + circle2.radius:
        return []
    # One circle within other
    if d_param < abs(circle1.radius - circle2.radius):
        return []
    # coincident circles
    if d_param == 0 and circle1.radius == circle2.radius:
        return []
    a = (circle1.radius ** 2 - circle2.radius ** 2 + d_param ** 2) / (2 * d_param)
    h_param = math.sqrt(circle1.radius ** 2 - a ** 2)
    x2 = x0 + a * (x1 - x0) / d_param
    y2 = y0 + a * (y1 - y0) / d_param
    x3 = x2 + h_param * (y1 - y0) / d_param
    y3 = y2 - h_param * (x1 - x0) / d_param

    x4 = x2 - h_param * (y1 - y0) / d_param
    y4 = y2 + h_param * (x1 - x0) / d_param

    return [volmdlr.Point2D(x3, y3), volmdlr.Point2D(x4, y4)]


def bspline_intersections_initial_conditions(primitive, bsplinecurve, resolution: float = 10):
    """
    Gets the initial conditions to calculate intersections between a bspline curve 2d and another edge 2d.

    :param primitive: primitive to verify intersection with bspline
    :param bsplinecurve: bsplinecurve to search for intersections.
    :param resolution: bspline discretization resolution, to search for initial intersection conditions.
    :return: a list with all initial sections where there may exist an intersection.
    """
    line_seg_class_ = getattr(volmdlr.edges, 'LineSegment'+bsplinecurve.__class__.__name__[-2:])
    abscissa1 = 0
    abscissa2 = bsplinecurve.length()
    bspline_discretized_points, points_abscissas = get_abscissa_discretization(bsplinecurve, abscissa1, abscissa2,
                                                                               max_number_points=resolution)
    if bsplinecurve.periodic:
        bspline_discretized_points += [bspline_discretized_points[0]]
        if points_abscissas[0] == 0.0:
            points_abscissas += [bsplinecurve.length()]
        else:
            points_abscissas += [points_abscissas[0]]
    param_intersections = []
    for point1, point2, abscissa1, abscissa2 in zip(bspline_discretized_points[:-1], bspline_discretized_points[1:],
                                                    points_abscissas[:-1], points_abscissas[1:]):
        line_seg = line_seg_class_(point1, point2)
        intersection = primitive.linesegment_intersections(line_seg)
        if intersection:
            param_intersections.append((abscissa1, abscissa2))
    if not param_intersections:
        param_intersections.append((0.0, bsplinecurve.length()))
    return param_intersections


def get_bsplinecurve_intersections(primitive, bsplinecurve, abs_tol: float = 1e-6):
    """
    Calculates the intersections between a primitive and a BSpline Curve.

    This method calculates the intersections between a given primitive and a BSpline Curve.
    The primitive can be any edge type or a plane3D. It first determines the initial intersection
    conditions using the `bspline_intersections_initial_conditions` function. Then, it iteratively
    checks for intersections by discretizing the BSpline Curve and checking for intersections with
    line segments. It maintains a list of intersections and validates each intersection against
    the specified tolerance. The method continues until there are no more intersection conditions
    to process. The resulting intersections are returned as a list.

    :param primitive: The primitive object to verify intersection with the BSpline Curve.
    It can be any edge type or a plane3D.
    :param bsplinecurve: The BSpline Curve object to search for intersections.
    :param abs_tol: The tolerance to be considered while validating an intersection.

    :return: A list with all intersections between the edge and BSpline Curve.
    :rtype: [volmdlr.Point3D].
    """
    param_intersections = bspline_intersections_initial_conditions(primitive, bsplinecurve)
    line_seg_class_ = getattr(volmdlr.edges, 'LineSegment' + bsplinecurve.__class__.__name__[-2:])
    intersections = []
    while True:
        if not param_intersections:
            break
        abscissa1, abscissa2 = param_intersections[0]
        discretized_points_between_1_2, points_abscissas = get_abscissa_discretization(bsplinecurve, abscissa1,
                                                                                       abscissa2, max_number_points=10)
        for point1, point2, abscissa_point1, abscissa_point2 in zip(
                discretized_points_between_1_2[:-1], discretized_points_between_1_2[1:],
                points_abscissas[:-1], points_abscissas[1:]):
            line_seg = line_seg_class_(point1, point2)
            intersection = primitive.linesegment_intersections(line_seg, 1e-6)
            if not intersection:
                continue
            if bsplinecurve.point_distance(intersection[0]) > 1e-6:
                param_intersections.insert(0, (abscissa_point1, abscissa_point2))
            elif not volmdlr.core.point_in_list(intersection[0], intersections):
                intersections.append(intersection[0])
        param_intersections.remove((abscissa1, abscissa2))
    return intersections


def conic_intersections(conic1, conic2, abs_tol: float = 1e-6):
    """
    Gets intersections between a two conic curves 3D.

    :param conic1: First conic curve 3D.
    :param conic2: Other conic curve 3D.
    :param abs_tol: tolerance.
    :return: A list of points, containing all intersections between the Line 3D and the Parabola3D.
    """
    intersections = []
    if conic1.frame.w.is_colinear_to(conic2.frame.w) and \
            math.isclose(conic1.frame.w.dot(conic2.frame.origin - conic1.frame.origin), 0, abs_tol=abs_tol):
        frame_mapped_conic1 = conic1.frame_mapping(conic1.frame, 'new')
        frame_mapped_conic2 = conic2.frame_mapping(conic1.frame, 'new')
        conic1_2d = frame_mapped_conic1.self_2d
        conic2_2d = frame_mapped_conic2.to_2d(frame_mapped_conic1.frame.origin,
                                              frame_mapped_conic1.frame.u,frame_mapped_conic1.frame.v)
        intersections_2d = conic1_2d.curve_intersections(conic2_2d, abs_tol)
        local_intersections = []
        for intersection in intersections_2d:
            local_intersections.append(volmdlr.Point3D(intersection[0], intersection[1], conic1.frame.origin.z))
        # circle_linseg_intersections = conic3d_line_intersections(frame_mapped_conic1, frame_mapped_conic2, abs_tol)
        for inter in local_intersections:
            intersections.append(conic1.frame.local_to_global_coordinates(inter))
        return intersections

    plane_intersections = get_two_planes_intersections(conic1.frame, conic2.frame)
    if not plane_intersections:
        return []
    plane_intersections = volmdlr.curves.Line3D(plane_intersections[0], plane_intersections[1])
    self_ellipse3d_line_intersections = conic3d_line_intersections(conic1, plane_intersections)
    ellipse3d_line_intersections = conic3d_line_intersections(conic2, plane_intersections)
    for intersection in self_ellipse3d_line_intersections + ellipse3d_line_intersections:
        if volmdlr.core.point_in_list(intersection, intersections):
            continue
        if conic1.point_belongs(intersection, abs_tol) and conic2.point_belongs(intersection, abs_tol):
            intersections.append(intersection)
    return intersections


def get_plane_linesegment_intersections(plane_frame, linesegment, abs_tol: float = 1e-6):
    """
    Gets the intersections of a plane a line segment 3d.

    :param plane_frame: the plane's frame.
    :param linesegment: other line segment.
    :param abs_tol: tolerance allowed.
    :return: a list with the intersecting point.
    """
    u_vector = linesegment.end - linesegment.start
    w_vector = linesegment.start - plane_frame.origin
    normaldotu = plane_frame.w.dot(u_vector)
    if normaldotu == 0.0 or math.isclose(plane_frame.w.unit_vector().dot(u_vector.unit_vector()),
                                         0.0, abs_tol=abs_tol):
        return []
    intersection_abscissea = - plane_frame.w.dot(w_vector) / normaldotu
    if intersection_abscissea < 0 or intersection_abscissea > 1:
        if math.isclose(abs(intersection_abscissea), 0, abs_tol=abs_tol):
            return [linesegment.start]
        if math.isclose(intersection_abscissea, 1, abs_tol=abs_tol):
            return [linesegment.end]
        return []
    return [linesegment.start + intersection_abscissea * u_vector]


def get_plane_line_intersections(plane_frame, line, abs_tol: float = 1e-8):
    """
    Find the intersection with a line.

    :param plane_frame: the plane's frame.
    :param line: Line to evaluate the intersection
    :type line: :class:`edges.Line`.
    :param abs_tol: tolerance.
    :type abs_tol: float.
    :return: ADD DESCRIPTION
    :rtype: List[volmdlr.Point3D]
    """
    u_vector = line.point2 - line.point1
    w_vector = line.point1 - plane_frame.origin
    if math.isclose(plane_frame.w.dot(u_vector), 0, abs_tol=abs_tol):
        return []
    intersection_abscissea = - plane_frame.w.dot(w_vector) / plane_frame.w.dot(u_vector)
    return [line.point1 + intersection_abscissea * u_vector]


def get_two_planes_intersections(plane1_frame, plane2_frame):
    """
    Calculates the intersections between two planes, given their frames.

    :param plane1_frame: Plane's 1 frame.
    :param plane2_frame: Plane's 2 frame.
    :return: A list containing two points that define an infinite line if there is any intersections,
    or an empty list if the planes are parallel.
    """
    if plane1_frame.w.is_colinear_to(plane2_frame.w):
        return []
    line_direction = plane1_frame.w.cross(plane2_frame.w)

    if line_direction.norm() < 1e-6:
        return None

    a1, b1, c1, d1 = get_plane_equation_coefficients(plane1_frame)
    a2, b2, c2, d2 = get_plane_equation_coefficients(plane2_frame)
    if not math.isclose(a1 * b2 - a2 * b1, 0.0, abs_tol=1e-10):
        x0 = (b1 * d2 - b2 * d1) / (a1 * b2 - a2 * b1)
        y0 = (a2 * d1 - a1 * d2) / (a1 * b2 - a2 * b1)
        point1 = volmdlr.Point3D(x0, y0, 0)
    elif a2 * c1 != a1 * c2:
        x0 = (c2 * d1 - c1 * d2) / (a2 * c1 - a1 * c2)
        z0 = (a1 * d2 - a2 * d1) / (a2 * c1 - a1 * c2)
        point1 = volmdlr.Point3D(x0, 0, z0)
    elif c1 * b2 != b1 * c2:
        y0 = (- c2 * d1 + c1 * d2) / (b1 * c2 - c1 * b2)
        z0 = (- b1 * d2 + b2 * d1) / (b1 * c2 - c1 * b2)
        point1 = volmdlr.Point3D(0, y0, z0)
    else:
        raise NotImplementedError
    return [point1, point1 + line_direction]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Common primitives 3D.
"""

import math
import warnings
from random import uniform
from typing import Dict, List, Tuple

import plot_data

import dessia_common.core as dc
import matplotlib.pyplot as plt
import numpy as npy
from scipy.optimize import Bounds, NonlinearConstraint, minimize
from scipy.stats import qmc

import volmdlr
import volmdlr.core
import volmdlr.edges
import volmdlr.faces
import volmdlr.primitives
import volmdlr.wires
from volmdlr import shells, surfaces, curves


npy.seterr(divide='raise')


class RoundedLineSegments3D(volmdlr.primitives.RoundedLineSegments):
    """
    A class representing a series of rounded line segments in 3D.

    This class inherits from the `RoundedLineSegments` class,
    and provides methods to work with rounded line segments in 3D.

    :param points: The list of points defining the line segments.
    :type points: List[volmdlr.Point3D]
    :param radius: The dictionary mapping segment indices to their respective radii.
    :type radius:Dict[int, float]
    :param adapt_radius: Flag indicating whether to adapt the radius based on segment length.
    Defaults to False.
    :type adapt_radius: bool, optional
    :param name: The name of the rounded line segments. Defaults to ''.
    :type name: str, optional
    """
    line_class = volmdlr.edges.LineSegment3D
    arc_class = volmdlr.edges.Arc3D

    def __init__(self, points: List[volmdlr.Point3D], radius: Dict[str, float],
                 adapt_radius: bool = False, name: str = ''):
        volmdlr.primitives.RoundedLineSegments.__init__(self, points, radius, adapt_radius=adapt_radius, name=name)

    def arc_features(self, point_index: int):
        radius = self.radius[point_index]
        point_1, point_i, point_2 = self.get_points(point_index)
        dist1 = (point_1 - point_i).norm()
        dist2 = (point_2 - point_i).norm()
        dist3 = (point_1 - point_2).norm()
        alpha = math.acos(-(dist3**2 - dist1**2 - dist2**2) / (2 * dist1 * dist2)) / 2.
        dist = radius / math.tan(alpha)

        u1 = (point_1 - point_i) / dist1
        u2 = (point_2 - point_i) / dist2

        p3 = point_i + u1 * dist
        p4 = point_i + u2 * dist

        n = u1.cross(u2)
        n /= n.norm()
        v1 = u1.cross(n)
        v2 = u2.cross(n)

        line1 = curves.Line3D(p3, p3 + v1)
        line2 = curves.Line3D(p4, p4 + v2)

        w = u1 + u2  # mean of v1 and v2
        w /= w.norm()

        interior = line1.minimum_distance_points(line2)[0] - w * radius
        return p3, interior, p4, dist, alpha

    def rotation(self, center: volmdlr.Point3D, axis: volmdlr.Vector3D,
                 angle: float):
        """
        OpenRoundedLineSegments3D rotation.

        :param center: rotation center
        :param axis: rotation axis
        :param angle: angle rotation
        :return: a new rotated OpenRoundedLineSegments3D
        """
        return self.__class__([point.rotation(center, axis, angle)
                               for point in self.points],
                              self.radius, self.closed, self.name)

    def translation(self, offset: volmdlr.Vector3D):
        """
        OpenRoundedLineSegments3D translation.

        :param offset: translation vector
        :return: A new translated OpenRoundedLineSegments3D
        """
        return self.__class__([point.translation(offset)
                               for point in self.points],
                              self.radius, self.closed, self.name)


class OpenRoundedLineSegments3D(volmdlr.wires.Wire3D, RoundedLineSegments3D):
    """
    Defines an open rounded line segments.

    :param points: Points used to draw the wire.
    :type points: List of Point3D.
    :param radius: Radius used to connect different parts of the wire.
    :type radius: {position1(n): float which is the radius linked the n-1 and.
    n+1 points, position2(n+1):...}
    """
    _non_data_eq_attributes = ['name']
    _non_data_hash_attributes = ['name']

    def __init__(self, points: List[volmdlr.Point3D], radius: Dict[str, float],
                 adapt_radius: bool = False, name: str = ''):
        RoundedLineSegments3D.__init__(self, points, radius, adapt_radius=adapt_radius, name='')
        self.closed = False

        volmdlr.wires.Wire3D.__init__(self, self._primitives(), name)


class ClosedRoundedLineSegments3D(RoundedLineSegments3D, volmdlr.wires.Contour3D):
    """
    Defines a closed rounded line segment in 3D.

    :param points: Points used to draw the wire
    :type points: List of Point3D
    :param radius: Radius used to connect different parts of the wire
    :type radius: {position1(n): float which is the radius linked the n-1 and
    n+1 points, position2(n+1):...}
    """
    _non_serializable_attributes = []
    _non_data_eq_attributes = ['name']
    _non_data_hash_attributes = ['name']

    def __init__(self, points: List[volmdlr.Point3D], radius: float, adapt_radius: bool = False, name: str = ''):
        RoundedLineSegments3D.__init__(self, points, radius, adapt_radius, name)
        self.closed = True
        volmdlr.wires.Contour3D.__init__(self, primitives=self._primitives(), name=name)


class Block(shells.ClosedShell3D):
    """
    Creates a block.

    :param frame: a frame 3D. The origin of the frame is the center of the block,
     the 3 vectors are defining the edges. The frame has not to be orthogonal
    """
    _standalone_in_db = True
    _non_serializable_attributes = ['size', 'bounding_box', 'faces', 'contours', 'plane', 'points', 'polygon2D']
    _non_data_eq_attributes = ['name', 'color', 'alpha', 'size', 'bounding_box',
                               'faces', 'contours', 'plane', 'points', 'polygon2D']
    _non_data_hash_attributes = []

    def __init__(self, frame: volmdlr.Frame3D, *,
                 color: Tuple[float, float, float] = None, alpha: float = 1.,
                 name: str = ''):
        self.frame = frame
        self.size = (self.frame.u.norm(),
                     self.frame.v.norm(),
                     self.frame.w.norm())
        self._octree = None
        self._quadtree = None
        faces = self.shell_faces()
        for face in faces:
            face.alpha = alpha
            face.color = color
        shells.ClosedShell3D.__init__(self, faces, color=color, alpha=alpha, name=name)

    def __eq__(self, other):
        if self.__class__.__name__ != other.__class__.__name__:
            return False
        if self.frame != other.frame:
            return False
        return True

    def __hash__(self):
        return hash((self.__class__.__name__, self.frame))

    def to_dict(self, *args, **kwargs):
        """
        Custom to_dict for performance.

        """
        dict_ = dc.DessiaObject.base_dict(self)
        dict_.update({'color': self.color,
                      'alpha': self.alpha,
                      'frame': self.frame.to_dict()})

        return dict_

    def volume(self):
        """Returns the volume of the block."""
        return self.size[0] * self.size[1] * self.size[2]

    @classmethod
    def from_bounding_box(cls, bounding_box, name: str = ''):
        """
        Transform a bounding box into a block.
        """
        origin = bounding_box.center
        bbox_size = bounding_box.size
        frame = volmdlr.Frame3D(origin, bbox_size[0] * volmdlr.Vector3D(1, 0, 0),
                                bbox_size[1] * volmdlr.Vector3D(0, 1, 0),
                                bbox_size[2] * volmdlr.Vector3D(0, 0, 1))
        return cls(frame=frame, name=name)

    def vertices(self):
        """Computes the vertices of the block."""
        return [self.frame.origin - 0.5 * self.frame.u - 0.5 * self.frame.v - 0.5 * self.frame.w,
                self.frame.origin - 0.5 * self.frame.u + 0.5 * self.frame.v - 0.5 * self.frame.w,
                self.frame.origin + 0.5 * self.frame.u + 0.5 * self.frame.v - 0.5 * self.frame.w,
                self.frame.origin + 0.5 * self.frame.u - 0.5 * self.frame.v - 0.5 * self.frame.w,
                self.frame.origin - 0.5 * self.frame.u - 0.5 * self.frame.v + 0.5 * self.frame.w,
                self.frame.origin - 0.5 * self.frame.u + 0.5 * self.frame.v + 0.5 * self.frame.w,
                self.frame.origin + 0.5 * self.frame.u + 0.5 * self.frame.v + 0.5 * self.frame.w,
                self.frame.origin + 0.5 * self.frame.u - 0.5 * self.frame.v + 0.5 * self.frame.w]

    def edges(self):
        """Computes the edges of the block."""
        point1, point2, point3, point4, point5, point6, point7, point8 = self.vertices()
        return [volmdlr.edges.LineSegment3D(point1.copy(), point2.copy()),
                volmdlr.edges.LineSegment3D(point2.copy(), point3.copy()),
                volmdlr.edges.LineSegment3D(point3.copy(), point4.copy()),
                volmdlr.edges.LineSegment3D(point4.copy(), point1.copy()),
                volmdlr.edges.LineSegment3D(point5.copy(), point6.copy()),
                volmdlr.edges.LineSegment3D(point6.copy(), point7.copy()),
                volmdlr.edges.LineSegment3D(point7.copy(), point8.copy()),
                volmdlr.edges.LineSegment3D(point8.copy(), point5.copy()),
                volmdlr.edges.LineSegment3D(point1.copy(), point5.copy()),
                volmdlr.edges.LineSegment3D(point2.copy(), point6.copy()),
                volmdlr.edges.LineSegment3D(point3.copy(), point7.copy()),
                volmdlr.edges.LineSegment3D(point4.copy(), point8.copy())]

    def face_contours3d(self):
        edge1, edge2, edge3, edge4, edge5, edge6, edge7, edge8, edge9, edge10, edge11, edge12 = self.edges()
        e5_switch = edge5.reverse()
        e6_switch = edge6.reverse()
        e7_switch = edge7.reverse()
        e8_switch = edge8.reverse()
        e9_switch = edge9.reverse()
        e10_switch = edge10.reverse()
        e11_switch = edge11.reverse()
        e12_switch = edge12.reverse()
        return [volmdlr.wires.Contour3D([edge1.copy(), edge2.copy(), edge3.copy(), edge4.copy()]),
                volmdlr.wires.Contour3D([edge5.copy(), edge6.copy(), edge7.copy(), edge8.copy()]),
                # volmdlr.Contour3D([e1.copy(), e9.copy(), e5.copy(), e10.copy()]),
                volmdlr.wires.Contour3D([edge1.copy(), edge10.copy(), e5_switch.copy(), e9_switch.copy()]),
                # volmdlr.Contour3D([e2.copy(), e10.copy(), e6.copy(), e11.copy()]),
                volmdlr.wires.Contour3D([edge2.copy(), edge11.copy(), e6_switch.copy(), e10_switch.copy()]),
                # volmdlr.Contour3D([e3.copy(), e11.copy(), e7.copy(), e12.copy()]),
                volmdlr.wires.Contour3D([edge3.copy(), edge12.copy(), e7_switch.copy(), e11_switch.copy()]),
                # volmdlr.Contour3D([e4.copy(), e12.copy(), e8.copy(), e9.copy()])]
                volmdlr.wires.Contour3D([edge4.copy(), edge9.copy(), e8_switch.copy(), e12_switch.copy()])]

    def shell_faces(self):
        """Computes the faces of the block."""
        hlx = 0.5 * self.frame.u.norm()
        hly = 0.5 * self.frame.v.norm()
        hlz = 0.5 * self.frame.w.norm()
        frame = self.frame.copy()
        frame = frame.normalize()
        xm_frame = volmdlr.Frame3D(frame.origin - 0.5 * self.frame.u,
                                   frame.v, frame.w, frame.u)
        xp_frame = volmdlr.Frame3D(frame.origin + 0.5 * self.frame.u,
                                   frame.v, frame.w, frame.u)
        ym_frame = volmdlr.Frame3D(frame.origin - 0.5 * self.frame.v,
                                   frame.w, frame.u, frame.v)
        yp_frame = volmdlr.Frame3D(frame.origin + 0.5 * self.frame.v,
                                   frame.w, frame.u, frame.v)
        zm_frame = volmdlr.Frame3D(frame.origin - 0.5 * self.frame.w,
                                   frame.u, frame.v, frame.w)
        zp_frame = volmdlr.Frame3D(frame.origin + 0.5 * self.frame.w,
                                   frame.u, frame.v, frame.w)

        xm_face = volmdlr.faces.PlaneFace3D.from_surface_rectangular_cut(
            surfaces.Plane3D(xm_frame), -hly, hly, -hlz, hlz)
        xp_face = volmdlr.faces.PlaneFace3D.from_surface_rectangular_cut(
            surfaces.Plane3D(xp_frame), -hly, hly, -hlz, hlz)
        ym_face = volmdlr.faces.PlaneFace3D.from_surface_rectangular_cut(
            surfaces.Plane3D(ym_frame), -hlz, hlz, -hlx, hlx)
        yp_face = volmdlr.faces.PlaneFace3D.from_surface_rectangular_cut(
            surfaces.Plane3D(yp_frame), -hlz, hlz, -hlx, hlx)
        zm_face = volmdlr.faces.PlaneFace3D.from_surface_rectangular_cut(
            surfaces.Plane3D(zm_frame), -hlx, hlx, -hly, hly)
        zp_face = volmdlr.faces.PlaneFace3D.from_surface_rectangular_cut(
            surfaces.Plane3D(zp_frame), -hlx, hlx, -hly, hly)

        return [xm_face, xp_face, ym_face, yp_face, zm_face, zp_face]

    def faces_center(self):
        """Computes the faces center of the block."""
        return [self.frame.origin - 0.5 * self.frame.u,
                self.frame.origin + 0.5 * self.frame.u,
                self.frame.origin - 0.5 * self.frame.v,
                self.frame.origin + 0.5 * self.frame.v,
                self.frame.origin - 0.5 * self.frame.w,
                self.frame.origin + 0.5 * self.frame.w]

    def rotation(self, center: volmdlr.Point3D, axis: volmdlr.Vector3D,
                 angle: float):
        """
        Block rotation.

        :param center: rotation center
        :param axis: rotation axis
        :param angle: angle rotation
        :return: a new rotated Block
        """
        new_frame = self.frame.rotation(center, axis, angle)
        return Block(new_frame, color=self.color,
                     alpha=self.alpha, name=self.name)

    def translation(self, offset: volmdlr.Vector3D):
        """
        Returns a new translated block.

        :param offset: translation vector.
        :return: A new translated Block.
        """
        new_frame = self.frame.translation(offset)
        return Block(new_frame, color=self.color,
                     alpha=self.alpha, name=self.name)

    def cut_by_orthogonal_plane(self, plane_3d: surfaces.Plane3D):
        bouding_box = self.bounding_box
        if plane_3d.frame.w.dot(volmdlr.Vector3D(1, 0, 0)) == 0:
            pass
        elif plane_3d.frame.w.dot(volmdlr.Vector3D(0, 1, 0)) == 0:
            pass
        elif plane_3d.frame.w.dot(volmdlr.Vector3D(0, 0, 1)) == 0:
            pass
        else:
            raise KeyError('plane is not orthogonal either with x, y or z')
        point_min = volmdlr.Point3D(bouding_box.xmin, bouding_box.ymin,
                                    bouding_box.zmin)
        point_max = volmdlr.Point3D(bouding_box.xmax, bouding_box.ymax,
                                    bouding_box.zmax)
        point_min_2d = plane_3d.point3d_to_2d(point_min)
        point_max_2d = plane_3d.point3d_to_2d(point_max)
        points = [point_min_2d, volmdlr.Point2D(point_max_2d.x, point_min_2d.y),
                  point_max_2d, volmdlr.Point2D(point_min_2d.x, point_max_2d.y)]
        contour_2d = surfaces.Surface2D(
            volmdlr.wires.ClosedPolygon2D(points), [])

        return volmdlr.faces.PlaneFace3D(plane_3d, contour_2d)

    def frame_mapping_parametres(self, frame: volmdlr.Frame3D,
                                 side: str):
        basis = frame.basis()
        if side == 'new':
            new_origin = frame.global_to_local_coordinates(self.frame.origin)
            new_u = basis.global_to_local_coordinates(self.frame.u)
            new_v = basis.global_to_local_coordinates(self.frame.v)
            new_w = basis.global_to_local_coordinates(self.frame.w)
            new_frame = volmdlr.Frame3D(new_origin, new_u, new_v, new_w)
        elif side == 'old':
            new_origin = frame.local_to_global_coordinates(self.frame.origin)
            new_u = basis.local_to_global_coordinates(self.frame.u)
            new_v = basis.local_to_global_coordinates(self.frame.v)
            new_w = basis.local_to_global_coordinates(self.frame.w)
            new_frame = volmdlr.Frame3D(new_origin, new_u, new_v, new_w)
        else:
            raise ValueError('side value not valid, please specify'
                             'a correct value: \'old\' or \'new\'')
        return new_frame

    def frame_mapping(self, frame: volmdlr.Frame3D, side: str):
        """
        Changes frame_mapping and return a new Frame3D.

        :param side: 'old' or 'new'
        """
        new_frame = self.frame_mapping_parametres(frame, side)
        return Block(new_frame, color=self.color,
                     alpha=self.alpha, name=self.name)

    def copy(self, deep=True, memo=None):
        """
        Creates a copy of a Block.

        """
        new_origin = self.frame.origin.copy()
        new_u = self.frame.u.copy()
        new_v = self.frame.v.copy()
        new_w = self.frame.w.copy()
        new_frame = volmdlr.Frame3D(new_origin, new_u, new_v, new_w)
        return Block(new_frame, color=self.color,
                     alpha=self.alpha, name=self.name)

    def plot_data(self, x3d, y3d, edge_style=plot_data.EdgeStyle):
        """Plot the 2D projections of a block."""
        lines = []
        for edge3d in self.edges():
            lines.append(edge3d.plot_data(x3d, y3d, edge_style))

        return lines

    def plot2d(self, x3d, y3d, ax=None):
        """
        Plot 2d with Matplotlib.

        """
        if ax is None:
            fig, ax = plt.subplots()
            ax.set_aspect('equal')
        else:
            fig = None

        for edge3d in self.edges():
            edge3d.plot2d(x3d, y3d, ax=ax)

        return fig, ax

    def octree(self):
        """Subdivide block into eight other blocks."""
        if not self._octree:
            self._octree = self.subdivide_block(2, 2, 2)
        return self._octree

    def quadtree(self):
        """Subdivide block into four other blocks."""
        if not self._quadtree:
            self._quadtree = self.subdivide_block(2, 2, 1)
        return self._quadtree

    def subdivide_block(self, number_blocks_x, number_blocks_y, number_blocks_z):
        """Divide block into sub blocks."""
        filling_boxes_size = [self.size[0] / number_blocks_x, self.size[1] / number_blocks_y,
                              self.size[2] / number_blocks_z]
        initial_frame_center = self.frame.origin.copy(deep=True).translation(
            -volmdlr.Vector3D(self.size[0] / 2 - filling_boxes_size[0] / 2,
                              self.size[1] / 2 - filling_boxes_size[1] / 2,
                              self.size[2] / 2 - filling_boxes_size[2] / 2))
        xyz = [volmdlr.Vector3D(filling_boxes_size[0], 0, 0), volmdlr.Vector3D(0, filling_boxes_size[1], 0),
               volmdlr.Vector3D(0, 0, filling_boxes_size[2])]

        dividing_blocks = []
        for z_box in range(number_blocks_z):
            for y_box in range(number_blocks_y):
                for x_box in range(number_blocks_x):
                    translation_vector = volmdlr.Vector3D(x_box * filling_boxes_size[0], y_box * filling_boxes_size[1],
                                                          z_box * filling_boxes_size[2])
                    origin_point = initial_frame_center.translation(translation_vector)
                    block = Block(frame=volmdlr.Frame3D(origin_point, *xyz))
                    dividing_blocks.append(block)
        return dividing_blocks


class ExtrudedProfile(shells.ClosedShell3D):
    """
    Extrude a profile given by outer and inner contours.

    TODO: In the future change to a frame and a surface2D and an extrusion vector.
    """
    _non_serializable_attributes = ['faces', 'inner_contours3d',
                                    'outer_contour3d']

    def __init__(self, frame: volmdlr.Frame3D,
                 outer_contour2d: volmdlr.wires.Contour2D,
                 inner_contours2d: List[volmdlr.wires.Contour2D],
                 extrusion_length: float,
                 color: Tuple[float, float, float] = None, alpha: float = 1.,
                 name: str = ''):
        self.frame = frame

        self.outer_contour2d = outer_contour2d
        self.outer_contour3d = outer_contour2d.to_3d(self.frame.origin, self.frame.u, self.frame.v)
        self.extrusion_length = extrusion_length
        self.inner_contours2d = inner_contours2d
        self.extrusion_vector = self.frame.w * extrusion_length
        self.inner_contours3d = []
        self.color = color

        bool_areas = []
        for contour in inner_contours2d:
            self.inner_contours3d.append(contour.to_3d(self.frame.origin, self.frame.u, self.frame.v))
            if contour.area() > outer_contour2d.area():
                bool_areas.append(True)
            else:
                bool_areas.append(False)
        if any(bool_areas):
            raise ValueError('At least one inner contour is not contained in outer_contour.')

        faces = self.shell_faces()

        shells.ClosedShell3D.__init__(self, faces, color=color,
                                      alpha=alpha, name=name)

    def to_dict(self, *args, **kwargs):
        """
        Serialize the ExtrudedProfile.

        """
        dict_ = dc.DessiaObject.base_dict(self)
        dict_.update({'color': self.color,
                      'alpha': self.alpha,
                      'frame': self.frame.to_dict(),
                      'outer_contour2d': self.outer_contour2d.to_dict(),
                      'inner_contours2d': [c.to_dict() for c in self.inner_contours2d],
                      'extrusion_length': self.extrusion_length,
                      })

        return dict_

    def copy(self, deep=True, memo=None):
        """
        Creates a copy of Extruded Profile.

        """
        return self.__class__(
            frame=self.frame.copy(),
            outer_contour2d=self.outer_contour2d.copy(),
            inner_contours2d=[c.copy() for c in self.inner_contours2d],
            extrusion_length=self.extrusion_length,
            color=self.color,
            alpha=self.alpha,
            name=self.name)

    def shell_faces(self):
        """
        Computes the shell faces from init data.

        """
        lower_plane = surfaces.Plane3D.from_plane_vectors(
            self.frame.origin, self.frame.u, self.frame.v)
        lower_face = volmdlr.faces.PlaneFace3D(
            lower_plane, surfaces.Surface2D(self.outer_contour2d,
                                            self.inner_contours2d))

        upper_face = lower_face.translation(self.extrusion_vector)
        lateral_faces = []
        for primitive in self.outer_contour3d.primitives:
            lateral_faces.extend(primitive.extrusion(self.extrusion_vector))

        for inner_contour in self.inner_contours3d:
            for primitive in inner_contour.primitives:
                lateral_faces.extend(primitive.extrusion(self.extrusion_vector))

        return [lower_face, upper_face] + lateral_faces

    def area(self):
        """Returns the area of the extruded 2D surface."""
        areas = self.outer_contour2d.area()
        areas -= sum(contour.area() for contour in self.inner_contours2d)
        return areas

    def volume(self):
        z = self.frame.w
        return self.area() * self.extrusion_vector.dot(z)

    def frame_mapping(self, frame: volmdlr.Frame3D, side: str):
        """
        Changes frame_mapping and return a new ExtrudeProfile.

        :param side: = 'old' or 'new'.
        """
        return ExtrudedProfile(
            self.frame.frame_mapping(frame, side),
            self.outer_contour2d, self.inner_contours2d,
            self.extrusion_length)

    def rotation(self, center: volmdlr.Point3D, axis: volmdlr.Vector3D,
                 angle: float):
        """
        Extruded Profile rotation.

        :param center: rotation center.
        :param axis: rotation axis.
        :param angle: angle rotation.
        :return: a new rotated ExtrudedProfile.
        """
        return self.__class__(
            frame=self.frame.rotation(center, axis, angle),
            outer_contour2d=self.outer_contour2d,
            inner_contours2d=self.inner_contours2d,
            extrusion_length=self.extrusion_length,
            color=self.color, alpha=self.alpha)

    def translation(self, offset: volmdlr.Vector3D):
        """
        Extruded Profile translation.

        :param offset: translation vector
        :return: A new translated ExtrudedProfile
        """
        return self.__class__(
            frame=self.frame.translation(offset),
            outer_contour2d=self.outer_contour2d,
            inner_contours2d=self.inner_contours2d,
            extrusion_length=self.extrusion_length,
            color=self.color, alpha=self.alpha)


class RevolvedProfile(shells.ClosedShell3D):
    """
    Revolve a 2D profile along an axis around a certain angle.

    """
    _non_serializable_attributes = ['faces', 'contour3D']

    def __init__(self, frame: volmdlr.Frame3D,
                 contour2d: volmdlr.wires.Contour2D,
                 axis_point: volmdlr.Point3D, axis: volmdlr.Vector3D,
                 angle: float = 2 * math.pi, *,
                 color: Tuple[float, float, float] = None, alpha: float = 1,
                 name: str = ''):
        if frame.w.cross(axis).is_close(volmdlr.Vector3D(0.0, 0.0, 0.0)):
            raise ValueError(f"The normal vector of the Revolution's contour frame should not be parallel \n"
                             f"to revolution axis. frame.w: {frame.w}; revolution_axis: {axis}")
        self.contour2d = contour2d
        self.axis_point = axis_point
        self.axis = axis
        self.angle = angle
        self.frame = frame
        self.contour3d = self.contour2d.to_3d(frame.origin, frame.u, frame.v)

        faces = self.shell_faces()
        shells.ClosedShell3D.__init__(self, faces, color=color,
                                      alpha=alpha, name=name)

    def __eq__(self, other):
        if not self.__class__.__name__ == other.__class__.__name__:
            return False
        for self_param, other_param in zip([self.frame,
                                            self.contour2d, self.axis_point, self.axis, self.angle],
                                           [other.frame,
                                            other.contour2d, other.axis_point, other.axis, other.angle]
                                           ):
            if not self_param == other_param:
                return False
        return True

    def to_dict(self, *args, **kwargs):
        """
        Custom to dict for performance.
        """
        dict_ = dc.DessiaObject.base_dict(self)
        dict_.update({'color': self.color,
                      'alpha': self.alpha,
                      'frame': self.frame,
                      'contour2d': self.contour2d.to_dict(),
                      'axis_point': self.axis_point.to_dict(),
                      'angle': self.angle,
                      'axis': self.axis.to_dict()
                      })

        return dict_

    def copy(self, deep=True, memo=None):
        """
        Creates a copy of Revolved-profile.

        """
        return self.__class__(frame=self.frame.copy(),
                              contour2d=self.contour2d.copy(deep=deep, memo=memo),
                              axis=self.axis.copy(), angle=self.angle,
                              axis_point=self.axis_point.copy(),
                              color=self.color, alpha=self.alpha,
                              name=self.name)

    def shell_faces(self):
        """
        Computes the shell faces from init data.

        """
        faces = []

        for edge in self.contour3d.primitives:
            faces.extend(edge.revolution(self.axis_point, self.axis, self.angle))

        if not math.isclose(self.angle, volmdlr.TWO_PI, abs_tol=1e-9):
            # Adding contours face to close
            plane1 = surfaces.Plane3D(self.frame)
            face1 = volmdlr.faces.PlaneFace3D(
                plane1, surfaces.Surface2D(self.contour2d, []))
            face2 = face1.rotation(self.axis_point, self.axis, self.angle)
            faces.append(face1)
            faces.append(face2)

        return faces

    def volume(self):
        """
        Volume from Guldin formulae.

        """
        point1 = self.axis_point.plane_projection3d(self.frame.origin, self.frame.u, self.frame.v)
        p1_2d = point1.to_2d(self.axis_point, self.frame.u, self.frame.v)
        p2_3d = self.axis_point + volmdlr.Point3D(self.axis.vector)
        p2_2d = p2_3d.to_2d(self.frame.origin, self.frame.u, self.frame.v)
        axis_2d = curves.Line2D(p1_2d, p2_2d)
        com = self.contour2d.center_of_mass()
        if com is not False:
            dist = axis_2d.point_distance(com)
            return self.angle * dist * self.contour2d.area()
        return 0.

    def rotation(self, center: volmdlr.Point3D, axis: volmdlr.Vector3D,
                 angle: float):
        """
        Revolved profile rotation.

        :param center: rotation center
        :type center: volmdlr.Point3D
        :param axis: rotation axis
        :type axis: volmdlr.Vector3D
        :param angle: angle rotation
        :type angle: float
        :return: a new rotated Revolved Profile
        :rtype: Revolved Profile
        """
        return self.__class__(
            frame=self.frame.rotation(center, axis, angle),
            contour2d=self.contour2d,
            axis_point=self.axis_point.rotation(center, axis, angle),
            axis=self.axis.rotation(center=volmdlr.O3D, axis=axis,
                                    angle=angle),
            angle=self.angle,
            color=self.color, alpha=self.alpha)

    def translation(self, offset: volmdlr.Vector3D):
        """
        Revolved Profile translation.

        :param offset: translation vector.
        :return: A new translated Revolved Profile.
        """
        return self.__class__(
            frame=self.frame.translation(offset),
            contour2d=self.contour2d,
            axis_point=self.axis_point.translation(offset),
            axis=self.axis,
            angle=self.angle,
            color=self.color, alpha=self.alpha)

    def frame_mapping_parameters(self, frame: volmdlr.Frame3D, side: str):
        """Apply transformation to object's parameters."""
        basis = frame.Basis()
        if side == 'old':
            axis = basis.local_to_global_coordinates(self.axis)
        elif side == 'new':
            axis = basis.global_to_local_coordinates(self.axis)
        else:
            raise ValueError('side must be either old or new')

        return axis

    def frame_mapping(self, frame: volmdlr.Frame3D, side: str):
        """
        Changes frame_mapping and return a new Revolved Profile.

        side = 'old' or 'new'
        """
        axis = self.frame_mapping_parameters(frame, side)
        return RevolvedProfile(
            self.frame.frame_mapping(frame, side),
            self.contour2d,
            self.axis_point.frame_mapping(frame, side),
            axis=axis, angle=self.angle)


class Cylinder(shells.ClosedShell3D):
    """
    Represents a 3D cylinder defined by its frame, radius, and length.
    """
    # pylint: disable=too-many-arguments

    def __init__(
        self,
        frame: volmdlr.Frame3D,
        radius: float,
        length: float,
        color: Tuple[float, float, float] = None,
        alpha: float = 1.0,
        name: str = "",
    ):
        """
        Initializes the Cylinder instance.

        The `Cylinder` class creates a cylinder with the specified radius and length, positioned using the given frame.
        The axis of revolution of the cylinder corresponds to the local z-axis (w-axis) of the provided frame.

        :param frame: The reference frame defining the position and orientation of the cylinder.
            The w-axis of the frame corresponds to the axis of revolution of the cylinder.
        :type frame: volmdlr.Frame3D
        :param radius: The radius of the cylinder.
        :type radius: float
        :param length: The length of the cylinder.
        :type length: float
        :param color: The color of the cylinder as an RGB tuple. Default is None.
        :type color: Tuple[float, float, float], optional
        :param alpha: The opacity of the cylinder (0.0 to 1.0). Default is 1.0.
        :type alpha: float, optional
        :param name: The name of the cylinder. Default is an empty string.
        :type name: str, optional
        """
        self.frame = frame
        self.position = frame.origin
        self.axis = frame.w
        self.radius = radius
        self.length = length

        faces = self.shell_faces()

        shells.ClosedShell3D.__init__(self, faces=faces, color=color, alpha=alpha, name=name)

    def shell_faces(self):
        """
        Computes the shell faces from init data.
        """
        surface3d = surfaces.CylindricalSurface3D(
            self.frame.translation(-self.frame.w * (self.length * 0.5)), self.radius
        )
        cylindrical_face = volmdlr.faces.CylindricalFace3D.from_surface_rectangular_cut(
            surface3d, 0, 2 * math.pi, 0, self.length
        )
        lower_plane = surfaces.Plane3D.from_plane_vectors(
            self.frame.origin.translation(-self.frame.w * (self.length * 0.5)), self.frame.u, self.frame.v
        )
        circle = volmdlr.curves.Circle2D(
            self.position.to_2d(self.frame.origin, self.frame.u, self.frame.v), self.radius
        )
        lower_face = volmdlr.faces.PlaneFace3D(
            lower_plane, surfaces.Surface2D(volmdlr.wires.Contour2D([volmdlr.edges.FullArc2D.from_curve(circle)]), [])
        )
        upper_face = lower_face.translation(self.frame.w * self.length)

        return [lower_face, cylindrical_face, upper_face]

    def get_bounding_box(self) -> volmdlr.core.BoundingBox:
        """
        Computes the bounding box of a cylinder.

        :return: The BoundingBox of the Cylinder.
        :rtype: :class:`volmdlr.core.BoundingBox`
        """
        # This was copied for HollowCylinder. Inheritance removed to avoid problems
        radius = self.radius

        point_a = self.position - self.length / 2 * self.axis
        point_b = self.position + self.length / 2 * self.axis

        dx2 = (point_a[0] - point_b[0])**2
        dy2 = (point_a[1] - point_b[1])**2
        dz2 = (point_a[2] - point_b[2])**2

        if point_a[0] > point_b[0]:
            point_a, point_b = point_b, point_a
        xmin = point_a[0] - (((dy2 + dz2) / (dx2 + dy2 + dz2))**0.5) * radius
        xmax = point_b[0] + (((dy2 + dz2) / (dx2 + dy2 + dz2))**0.5) * radius

        if point_a[1] > point_b[1]:
            point_a, point_b = point_b, point_a
        ymin = point_a[1] - (((dx2 + dz2) / (dx2 + dy2 + dz2))**0.5) * radius
        ymax = point_b[1] + (((dx2 + dz2) / (dx2 + dy2 + dz2))**0.5) * radius

        if point_a[2] > point_b[2]:
            point_a, point_b = point_b, point_a
        zmin = point_a[2] - (((dx2 + dy2) / (dx2 + dy2 + dz2))**0.5) * radius
        zmax = point_b[2] + (((dx2 + dy2) / (dx2 + dy2 + dz2))**0.5) * radius

        return volmdlr.core.BoundingBox(xmin, xmax, ymin, ymax, zmin, zmax)

    def volume(self) -> float:
        """
        Compute the volume of the cylinder.

        :return: The computed volume of the Cylinder.
        :rtype: float
        """
        return self.length * math.pi * self.radius**2

    @classmethod
    def from_end_points(
        cls,
        point1: volmdlr.Point3D,
        point2: volmdlr.Point3D,
        radius: float,
        color: Tuple[float, float, float] = None,
        alpha: float = 1,
        name: str = "",
    ):
        """
        Create a cylinder from two end points.

        :param point1: The first end point defining the base of the cylinder.
        :type point1: volmdlr.Point3D
        :param point2: The second end point defining the top of the cylinder.
        :type point2: volmdlr.Point3D
        :param radius: The radius of the cylinder.
        :type radius: float
        :param color: The color of the cylinder as an RGB tuple. Default is None.
        :type color: Tuple[float, float, float], optional
        :param alpha: The opacity of the cylinder (0.0 to 1.0). Default is 1.0.
        :type alpha: float, optional
        :param name: The name of the cylinder. Default is an empty string.
        :type name: str, optional

        :return: A Cylinder instance created from the specified end points.
        :rtype: Cylinder
        """
        position = 0.5 * (point1 + point2)
        length = point1.point_distance(point2)
        axis = (point2 - point1).to_vector()
        axis = axis.unit_vector()
        u_vector = axis.deterministic_unit_normal_vector()
        v_vector = axis.cross(u_vector)

        frame = volmdlr.Frame3D(position, u_vector, v_vector, axis)

        return cls(frame=frame, radius=radius, length=length, color=color, alpha=alpha, name=name)

    @classmethod
    def from_extremal_points(
        cls,
        point1: volmdlr.Point3D,
        point2: volmdlr.Point3D,
        radius: float,
        color: Tuple[float, float, float] = None,
        alpha: float = 1,
        name: str = "",
    ):
        """Deprecated class method. Use 'from_end_points' instead."""
        warnings.warn("Deprecated classmethod. Use 'from_end_points' instead.", DeprecationWarning)

        return cls.from_end_points(point1, point2, radius, color, alpha, name)

    @classmethod
    def from_center_point_and_axis(
        cls,
        center_point: volmdlr.Point3D,
        axis: volmdlr.Vector3D,
        radius: float,
        length: float,
        color: Tuple[float, float, float] = None,
        alpha: float = 1,
        name: str = "",
    ) -> 'Cylinder':
        """
        Create a cylinder from a center point, an axis, radius, and length.

        :param center_point: The center point of the cylinder (i.e. the middle point of the axis of the cylinder).
        :type center_point: volmdlr.Point3D
        :param axis: The axis of revolution for the cylinder.
        :type axis: volmdlr.Vector3D
        :param radius: The radius of the cylinder.
        :type radius: float
        :param length: The length of the cylinder.
        :type length: float
        :param color: The color of the cylinder as an RGB tuple. Default is None.
        :type color: Tuple[float, float, float], optional
        :param alpha: The opacity of the cylinder (0.0 to 1.0). Default is 1.0.
        :type alpha: float, optional
        :param name: The name of the cylinder. Default is an empty string.
        :type name: str, optional

        :return: A Cylinder instance created from the specified center point, axis, radius, and length.
        :rtype: Cylinder
        """
        u_vector = axis.deterministic_unit_normal_vector()
        v_vector = axis.cross(u_vector)
        frame = volmdlr.Frame3D(center_point, u_vector, v_vector, axis)

        return cls(frame=frame, radius=radius, length=length, color=color, alpha=alpha, name=name)

    def rotation(self, center: volmdlr.Point3D, axis: volmdlr.Vector3D, angle: float) -> 'Cylinder':
        """
        Cylinder rotation.

        :param center: The rotation center.
        :type center: volmdlr.Point3D
        :param axis: The rotation axis.
        :type axis: volmdlr.Vector3D
        :param angle: The angle of rotation.
        :type angle: float

        :return: A new rotated Cylinder.
        :rtype: Cylinder
        """
        return self.__class__(
            frame=self.frame.rotation(center, axis, angle),
            length=self.length,
            radius=self.radius,
            color=self.color,
            alpha=self.alpha,
            name=self.name,
        )

    def translation(self, offset: volmdlr.Vector3D) -> 'Cylinder':
        """
        Cylinder translation.

        :param offset: The translation vector.
        :type offset: volmdlr.Vector3D

        :return: A new translated Cylinder.
        :rtype: Cylinder
        """
        return self.__class__(
            frame=self.frame.translation(offset),
            length=self.length,
            radius=self.radius,
            color=self.color,
            alpha=self.alpha,
            name=self.name,
        )

    def frame_mapping(self, frame: volmdlr.Frame3D, side: str) -> 'Cylinder':
        """
        Changes frame_mapping and return a new Frame3D.

        side = 'old' or 'new'
        """
        return Cylinder(
            frame=self.frame.frame_mapping(frame, side),
            radius=self.radius,
            length=self.length,
            color=self.color,
            alpha=self.alpha,
            name=self.name,
        )

    def copy(self, deep=True, memo=None) -> 'Cylinder':
        """
        Creates a copy of Cylinder.

        :return: A copy of a current Cylinder.
        :rtype: Cylinder
        """
        return Cylinder(
            frame=self.frame.copy(),
            radius=self.radius,
            length=self.length,
            color=self.color,
            alpha=self.alpha,
            name=self.name,
        )

    def min_distance_to_other_cylinder(self, other_cylinder: 'Cylinder') -> float:
        """
        Compute the minimal distance between two volmdlr cylinders.

        :param other_cylinder: The other cylinder to compute the distance with.
        :type other_cylinder: Cylinder

        :return: The minimal distance between the two 3D cylinders.
        :rtype: float
        """
        # Basic check
        if self.point_belongs(other_cylinder.position) or other_cylinder.point_belongs(self.position):
            return 0.

        # Local frames of cylinders
        frame0 = volmdlr.Frame3D.from_point_and_vector(
            point=self.position, vector=self.axis, main_axis=volmdlr.Z3D
        )
        frame1 = volmdlr.Frame3D.from_point_and_vector(
            point=other_cylinder.position,
            vector=other_cylinder.axis,
            main_axis=volmdlr.Z3D,
        )

        matrix0 = frame0.transfer_matrix()
        x0, y0, z0 = frame0.origin.x, frame0.origin.y, frame0.origin.z
        matrix1 = frame1.transfer_matrix()
        x1, y1, z1 = frame1.origin.x, frame1.origin.y, frame1.origin.z

        # Euclidean distance
        def dist(point0, point1):
            return math.sqrt(
                (point0[0] - point1[0]) ** 2 + (point0[1] - point1[1]) ** 2 + (point0[2] - point1[2]) ** 2
            )

        # Local coordinates to global coordinates
        def to_global_point(point, matrix, origin):
            return [
                matrix.M11 * point[0] + matrix.M12 * point[1] + matrix.M13 * point[2] + origin[0],
                matrix.M21 * point[0] + matrix.M22 * point[1] + matrix.M23 * point[2] + origin[1],
                matrix.M31 * point[0] + matrix.M32 * point[1] + matrix.M33 * point[2] + origin[2],
            ]

        # Objective function
        def objective(x_param):
            point0_ = to_global_point(x_param[:3], matrix0, [x0, y0, z0])
            point1_ = to_global_point(x_param[3:], matrix1, [x1, y1, z1])

            return dist(point0_, point1_)

        # Gradient of objective function
        def gradient_objective(x_param):
            point0_ = to_global_point(x_param[:3], matrix0, [x0, y0, z0])
            point1_ = to_global_point(x_param[3:], matrix1, [x1, y1, z1])

            distance = dist(point0_, point1_)

            return [
                (point0_[0] - point1_[0]) / distance * matrix0.M11
                + (point0_[1] - point1_[1]) / distance * matrix0.M21
                + (point0_[2] - point1_[2]) / distance * matrix0.M31,
                (point0_[0] - point1_[0]) / distance * matrix0.M12
                + (point0_[1] - point1_[1]) / distance * matrix0.M22
                + (point0_[2] - point1_[2]) / distance * matrix0.M32,
                (point0_[0] - point1_[0]) / distance * matrix0.M13
                + (point0_[1] - point1_[1]) / distance * matrix0.M23
                + (point0_[2] - point1_[2]) / distance * matrix0.M33,
                (point1_[0] - point0_[0]) / distance * matrix1.M11
                + (point1_[1] - point0_[1]) / distance * matrix1.M21
                + (point1_[2] - point0_[2]) / distance * matrix1.M31,
                (point1_[0] - point0_[0]) / distance * matrix1.M12
                + (point1_[1] - point0_[1]) / distance * matrix1.M22
                + (point1_[2] - point0_[2]) / distance * matrix1.M32,
                (point1_[0] - point0_[0]) / distance * matrix1.M13
                + (point1_[1] - point0_[1]) / distance * matrix1.M23
                + (point1_[2] - point0_[2]) / distance * matrix1.M33,
            ]

        # Initial vector
        initial_guess = npy.zeros(6)

        # Constraints
        def constraint_radius_0(x):
            # radius of cylinder 0
            return x[0] ** 2 + x[1] ** 2

        def constraint_radius_1(x):
            # radius of cylinder 1
            return x[3] ** 2 + x[4] ** 2

        def gradient_constraint_radius_0(x):
            # gradient of constraint_radius_0
            return [2 * x[0], 2 * x[1], 0, 0, 0, 0]

        def gradient_constraint_radius_1(x):
            # gradient of constraint_radius_1
            return [0, 0, 0, 2 * x[3], 2 * x[4], 0]

        constraints = [
            NonlinearConstraint(
                fun=constraint_radius_0,
                lb=0,
                ub=self.radius**2,
                jac=gradient_constraint_radius_0,
            ),
            NonlinearConstraint(
                fun=constraint_radius_1,
                lb=0,
                ub=other_cylinder.radius**2,
                jac=gradient_constraint_radius_1,
            ),
        ]

        # Bounds
        bounds = Bounds(
            lb=[
                -self.radius,
                -self.radius,
                -self.length / 2,
                -other_cylinder.radius,
                -other_cylinder.radius,
                -other_cylinder.length / 2,
            ],
            ub=[
                self.radius,
                self.radius,
                self.length / 2,
                other_cylinder.radius,
                other_cylinder.radius,
                other_cylinder.length / 2,
            ],
        )

        return minimize(
            fun=objective,
            x0=initial_guess,
            bounds=bounds,
            tol=1e-6,
            constraints=constraints,
            jac=gradient_objective,
        ).fun

    def is_intersecting_other_cylinder(self, other_cylinder: 'Cylinder') -> bool:
        """
        Verifies if two cylinders are intersecting or not.

        :param other_cylinder: The other cylinder to compute if there is an intersection with.
        :type other_cylinder: Cylinder

        :return: True if cylinders are intersecting, False otherwise
        :rtype: bool
        """
        dist = self.min_distance_to_other_cylinder(other_cylinder)

        return dist < 1e-5

    def random_point_inside(self) -> volmdlr.Point3D:
        """
        Return a random point inside a cylinder.

        :return: a random point inside the Cylinder
        :rtype: volmdlr.Point3D
        """
        theta = uniform(0, 2 * math.pi)
        radius = math.sqrt(uniform(0, 1)) * self.radius

        x_local = radius * math.cos(theta)
        y_local = radius * math.sin(theta)
        z_local = uniform(-self.length / 2, self.length / 2)

        local_frame = volmdlr.Frame3D.from_point_and_vector(
            point=self.position, vector=self.axis, main_axis=volmdlr.Z3D
        )

        return local_frame.local_to_global_coordinates(volmdlr.Point3D(x_local, y_local, z_local)).to_point()

    def lhs_points_inside(self, n_points: int) -> List[volmdlr.Point3D]:
        """
        Returns some points inside the cylinder from a LHS samplings.

        :param n_points: The number of points to generate.
        :type n_points: int

        :return: The Latin Hypercube Sampling points inside the cylinder.
        :rtype: list[volmdlr.Point3D]
        """
        local_frame = volmdlr.Frame3D.from_point_and_vector(
            point=self.position, vector=self.axis, main_axis=volmdlr.Z3D
        )

        # sampling point in Cartesian local coordinates
        sampler = qmc.LatinHypercube(d=3, seed=0)
        sample = qmc.scale(
            sampler.random(n=n_points),
            [0, 0, -self.length / 2],
            [1, 2 * math.pi, self.length / 2],
        )

        # converting sampled point in global coordinates volmdlr.Point3D points
        points = []
        for point in sample:
            radius = math.sqrt(point[0]) * self.radius
            theta = point[1]

            x_local = radius * math.cos(theta)
            y_local = radius * math.sin(theta)
            z_local = point[2]

            points.append(
                local_frame.local_to_global_coordinates(volmdlr.Point3D(x_local, y_local, z_local)).to_point()
            )

        return points

    def point_belongs(self, point3d: volmdlr.Point3D, **kwargs) -> bool:
        """
        Check if the point belongs to the cylinder.

        :param point3d: The point to check if it's the cylinder.
        :type point3d: volmdlr.Point3D

        :return: True if the given point is inside the cylinder, False otherwise.
        :rtype: bool
        """
        local_frame = volmdlr.Frame3D.from_point_and_vector(
            point=self.position, vector=self.axis, main_axis=volmdlr.Z3D
        )

        local_point = local_frame.global_to_local_coordinates(point3d)

        return (math.sqrt(local_point.x ** 2 + local_point.y ** 2) <= self.radius) and (
                -self.length / 2 <= local_point.z <= self.length / 2
        )

    def interference_volume_with_other_cylinder(self, other_cylinder: "Cylinder", n_points: int = 1000) -> float:
        """
        Estimation of the interpenetration volume using LHS sampling (inspired by Monte-Carlo method).

        :param other_cylinder: The other cylinder to compute the interference volume with.
        :type other_cylinder: Cylinder
        :param n_points: Optional parameter used for the number of random point used to discretize the cylinder
        :type n_points: int

        :return: An estimation of the interference volume.
        :rtype: float
        """

        # doing the discretization on the smallest cylinder to have better precision
        if self.volume() < other_cylinder.volume():
            smallest_cylinder = self
        else:
            smallest_cylinder = other_cylinder
            other_cylinder = self

        return (
                len(
                    [
                        point
                        for point in smallest_cylinder.lhs_points_inside(n_points)
                        if other_cylinder.point_belongs(point)
                    ]
                )
                / n_points
        ) * smallest_cylinder.volume()


class Cone(shells.ClosedShell3D):
    """
    Represents a 3D cone defined by its frame, radius, and length.
    """
    # pylint: disable=too-many-arguments

    def __init__(
        self,
        frame: volmdlr.Frame3D,
        radius: float,
        length: float,
        color: Tuple[float, float, float] = None,
        alpha: float = 1.0,
        name: str = "",
    ):
        """
        Initializes the Cone instance.

        The `Cone` class creates a cone with the specified radius and length, positioned using the given frame.
        The axis of revolution of the cone corresponds to the local z-axis (w-axis) of the provided frame.
        The top of the cone is oriented according to the local z-axis (w-axis) of the provided frame.
        The center of the cone axis is positioned at the provided frame origin point.

        :param frame: The reference frame defining the position and orientation of the cone.
            The w-axis of the frame corresponds to the axis of revolution of the cone.
        :type frame: volmdlr.Frame3D
        :param radius: The radius of the cone.
        :type radius: float
        :param length: The length of the cone.
        :type length: float
        :param color: The color of the cone as an RGB tuple. Default is None.
        :type color: Tuple[float, float, float], optional
        :param alpha: The opacity of the cone (0.0 to 1.0). Default is 1.0.
        :type alpha: float, optional
        :param name: The name of the cone. Default is an empty string.
        :type name: str, optional
        """
        self.frame = frame
        self.position = frame.origin
        self.axis = frame.w
        self.radius = radius
        self.length = length

        faces = self.shell_faces()

        shells.ClosedShell3D.__init__(self, faces=faces, color=color, alpha=alpha, name=name)

    def shell_faces(self):
        """
        Computes the shell faces from init data.
        """
        conical_surface_frame = self.frame.translation(self.frame.w * (self.length * 0.5))
        conical_surface_frame.w = -conical_surface_frame.w
        surface3d = surfaces.ConicalSurface3D(conical_surface_frame, math.atan(self.radius / self.length))

        conical_face = volmdlr.faces.ConicalFace3D.from_surface_rectangular_cut(
            surface3d, 0, 2 * math.pi, 0, self.length
        )
        lower_plane = surfaces.Plane3D.from_plane_vectors(
            self.frame.origin.translation(-self.frame.w * (self.length * 0.5)), self.frame.u, self.frame.v
        )
        circle = volmdlr.curves.Circle2D(
            self.position.to_2d(self.frame.origin, self.frame.u, self.frame.v), self.radius
        )
        lower_face = volmdlr.faces.PlaneFace3D(
            lower_plane, surfaces.Surface2D(volmdlr.wires.Contour2D([volmdlr.edges.FullArc2D.from_curve(circle)]), [])
        )

        return [lower_face, conical_face]

    def get_bounding_box(self) -> volmdlr.core.BoundingBox:
        """
        Compute the bounding box of the cone.

        A is the point at the basis.
        B is the top.

        :return: The BoundingBox of the Cone.
        :rtype: :class:`volmdlr.core.BoundingBox`
        """
        point_a = self.position - self.length / 2 * self.axis
        point_b = self.position + self.length / 2 * self.axis

        dx2 = (point_a[0] - point_b[0])**2
        dy2 = (point_a[1] - point_b[1])**2
        dz2 = (point_a[2] - point_b[2])**2

        x_bound = (point_a[0] - (((dy2 + dz2) / (dx2 + dy2 + dz2))**0.5) * self.radius,
                   point_a[0] + (((dy2 + dz2) / (dx2 + dy2 + dz2))**0.5) * self.radius, point_b[0])
        xmin = min(x_bound)
        xmax = max(x_bound)

        y_bound = (point_a[1] - (((dx2 + dz2) / (dx2 + dy2 + dz2))**0.5) * self.radius,
                   point_a[1] + (((dx2 + dz2) / (dx2 + dy2 + dz2))**0.5) * self.radius, point_b[1])
        ymin = min(y_bound)
        ymax = max(y_bound)

        z_bound = (point_a[2] - (((dx2 + dy2) / (dx2 + dy2 + dz2))**0.5) * self.radius,
                   point_a[2] + (((dx2 + dy2) / (dx2 + dy2 + dz2))**0.5) * self.radius, point_b[2])
        zmin = min(z_bound)
        zmax = max(z_bound)

        return volmdlr.core.BoundingBox(xmin, xmax, ymin, ymax, zmin, zmax)

    def rotation(self, center: volmdlr.Point3D, axis: volmdlr.Vector3D, angle: float) -> 'Cone':
        """
        Cone rotation.

        :param center: The rotation center.
        :type center: volmdlr.Point3D
        :param axis: The rotation axis.
        :type axis: volmdlr.Vector3D
        :param angle: The angle of rotation.
        :type angle: float

        :return: A new rotated Cone.
        :rtype: Cone
        """
        return self.__class__(
            frame=self.frame.rotation(center, axis, angle),
            radius=self.radius,
            length=self.length,
            color=self.color,
            alpha=self.alpha,
        )

    def translation(self, offset: volmdlr.Vector3D) -> 'Cone':
        """
        Cone translation.

        :param offset: The translation vector.
        :type offset: volmdlr.Vector3D

        :return: A new translated Cone.
        :rtype: Cone
        """
        return self.__class__(
            frame=self.frame.translation(offset),
            radius=self.radius,
            length=self.length,
            color=self.color,
            alpha=self.alpha,
        )

    def volume(self) -> float:
        """
        Compute the volume of the cone.

        :return: The computed volume of the cone.
        :rtype: float
        """
        return self.length * math.pi * self.radius**2 / 3

    @classmethod
    def from_center_point_and_axis(
        cls,
        center_point: volmdlr.Point3D,
        axis: volmdlr.Vector3D,
        radius: float,
        length: float,
        color: Tuple[float, float, float] = None,
        alpha: float = 1,
        name: str = "",
    ) -> 'Cone':
        """
        Create a cone from a center point, an axis, radius, and length.

        :param center_point: The center point of the cone (i.e. the middle point of the axis of the cone).
        :type center_point: volmdlr.Point3D
        :param axis: The axis of revolution for the cone.
        :type axis: volmdlr.Vector3D
        :param radius: The radius of the cone.
        :type radius: float
        :param length: The length of the cone.
        :type length: float
        :param color: The color of the cone as an RGB tuple. Default is None.
        :type color: Tuple[float, float, float], optional
        :param alpha: The opacity of the cone (0.0 to 1.0). Default is 1.0.
        :type alpha: float, optional
        :param name: The name of the cone. Default is an empty string.
        :type name: str, optional

        :return: A Cone instance created from the specified center point, axis, radius, and length.
        :rtype: Cone
        """
        u_vector = axis.deterministic_unit_normal_vector()
        v_vector = axis.cross(u_vector)
        frame = volmdlr.Frame3D(center_point, u_vector, v_vector, axis)

        return cls(frame=frame, radius=radius, length=length, color=color, alpha=alpha, name=name)


class HollowCylinder(shells.ClosedShell3D):
    """
    Represents a 3D hollow cylinder defined by its frame, radii, and length.
    """
    # pylint: disable=too-many-arguments

    def __init__(
        self,
        frame: volmdlr.Frame3D,
        inner_radius: float,
        outer_radius: float,
        length: float,
        color: Tuple[float, float, float] = None,
        alpha: float = 1,
        name: str = "",
    ):
        """
        Initializes the HollowCylinder instance.

        The `HollowCylinder` class creates a hollow cylinder with the specified radii and length, positioned using the
        given frame.
        The axis of revolution of the hollow cylinder corresponds to the local z-axis (w-axis) of the provided frame.

        :param frame: The reference frame defining the position and orientation of the hollow cylinder.
            The w-axis of the frame corresponds to the axis of revolution of the hollow cylinder.
        :type frame: volmdlr.Frame3D
        :param inner_radius: The inner radius of the hollow cylinder.
        :type inner_radius: float
        :param outer_radius: The outer radius of the hollow cylinder.
        :type outer_radius: float
        :param length: The length of the hollow cylinder.
        :type length: float
        :param color: The color of the hollow cylinder as an RGB tuple. Default is None.
        :type color: Tuple[float, float, float], optional
        :param alpha: The opacity of the hollow cylinder (0.0 to 1.0). Default is 1.0.
        :type alpha: float, optional
        :param name: The name of the hollow cylinder. Default is an empty string.
        :type name: str, optional
        """
        self.frame = frame
        self.position = frame.origin
        self.axis = frame.w
        self.inner_radius = inner_radius
        self.outer_radius = outer_radius
        self.length = length

        faces = self.shell_faces()
        shells.ClosedShell3D.__init__(self, faces=faces, color=color, alpha=alpha, name=name)

    def shell_faces(self):
        """
        Computes the shell faces from init data.
        """
        surface3d_1 = surfaces.CylindricalSurface3D(
            self.frame.translation(-self.frame.w * (self.length * 0.5)), self.outer_radius
        )
        surface3d_2 = surfaces.CylindricalSurface3D(
            self.frame.translation(-self.frame.w * (self.length * 0.5)), self.inner_radius
        )

        cylindrical_face1 = volmdlr.faces.CylindricalFace3D.from_surface_rectangular_cut(
            surface3d_1, 0, 2 * math.pi, 0, self.length
        )
        cylindrical_face2 = volmdlr.faces.CylindricalFace3D.from_surface_rectangular_cut(
            surface3d_2, 0, 2 * math.pi, 0, self.length
        )

        lower_plane = surfaces.Plane3D.from_plane_vectors(
            self.frame.origin.translation(-self.frame.w * (self.length * 0.5)), self.frame.u, self.frame.v
        )

        position_2d = self.position.to_2d(self.frame.origin, self.frame.u, self.frame.v)
        outer_circle = volmdlr.curves.Circle2D(position_2d, self.outer_radius)
        inner_circle = volmdlr.curves.Circle2D(position_2d, self.inner_radius)

        lower_face = volmdlr.faces.PlaneFace3D(
            lower_plane,
            surfaces.Surface2D(
                volmdlr.wires.Contour2D([volmdlr.edges.FullArc2D.from_curve(outer_circle)]),
                [volmdlr.wires.Contour2D([volmdlr.edges.FullArc2D.from_curve(inner_circle)])],
            ),
        )
        upper_face = lower_face.translation(self.frame.w * self.length)

        return [lower_face, cylindrical_face1, cylindrical_face2, upper_face]

    def get_bounding_box(self) -> volmdlr.core.BoundingBox:
        """
        Computes the bounding box of a hollow cylinder.

        :return: The BoundingBox of the HollowCylinder.
        :rtype: :class:`volmdlr.core.BoundingBox`
        """
        radius = self.outer_radius

        point_a = self.position - self.length / 2 * self.axis
        point_b = self.position + self.length / 2 * self.axis

        dx2 = (point_a[0] - point_b[0])**2
        dy2 = (point_a[1] - point_b[1])**2
        dz2 = (point_a[2] - point_b[2])**2

        if point_a[0] > point_b[0]:
            point_a, point_b = point_b, point_a
        xmin = point_a[0] - (((dy2 + dz2) / (dx2 + dy2 + dz2))**0.5) * radius
        xmax = point_b[0] + (((dy2 + dz2) / (dx2 + dy2 + dz2))**0.5) * radius

        if point_a[1] > point_b[1]:
            point_a, point_b = point_b, point_a
        ymin = point_a[1] - (((dx2 + dz2) / (dx2 + dy2 + dz2))**0.5) * radius
        ymax = point_b[1] + (((dx2 + dz2) / (dx2 + dy2 + dz2))**0.5) * radius

        if point_a[2] > point_b[2]:
            point_a, point_b = point_b, point_a
        zmin = point_a[2] - (((dx2 + dy2) / (dx2 + dy2 + dz2))**0.5) * radius
        zmax = point_b[2] + (((dx2 + dy2) / (dx2 + dy2 + dz2))**0.5) * radius

        return volmdlr.core.BoundingBox(xmin, xmax, ymin, ymax, zmin, zmax)

    def volume(self) -> float:
        """
        Compute the volume of the hollow cylinder.

        :return: The computed volume of the Cylinder.
        :rtype: float
        """
        return self.length * math.pi * (self.outer_radius**2 - self.inner_radius**2)

    @classmethod
    def from_end_points(
        cls,
        point1: volmdlr.Point3D,
        point2: volmdlr.Point3D,
        inner_radius: float,
        outer_radius: float,
        color: Tuple[float, float, float] = None,
        alpha: float = 1,
        name: str = "",
    ):
        """
        Create a hollow cylinder from two end points.

        :param point1: The first end point defining the base of the hollow cylinder.
        :type point1: volmdlr.Point3D
        :param point2: The second end point defining the top of the hollow cylinder.
        :type point2: volmdlr.Point3D
        :param inner_radius: The inner radius of the hollow cylinder.
        :type inner_radius: float
        :param outer_radius: The outer radius of the hollow cylinder.
        :type outer_radius: float
        :param color: The color of the hollow cylinder as an RGB tuple. Default is None.
        :type color: Tuple[float, float, float], optional
        :param alpha: The opacity of the hollow cylinder (0.0 to 1.0). Default is 1.0.
        :type alpha: float, optional
        :param name: The name of the hollow cylinder. Default is an empty string.
        :type name: str, optional

        :return: A HollowCylinder instance created from the specified end points.
        :rtype: HollowCylinder
        """
        position = 0.5 * (point1 + point2)
        length = point1.point_distance(point2)
        axis = (point2 - point1).to_vector()
        axis = axis.unit_vector()
        u_vector = axis.deterministic_unit_normal_vector()
        v_vector = axis.cross(u_vector)

        frame = volmdlr.Frame3D(position, u_vector, v_vector, axis)

        return cls(
            frame=frame,
            inner_radius=inner_radius,
            outer_radius=outer_radius,
            length=length,
            color=color,
            alpha=alpha,
            name=name,
        )

    @classmethod
    def from_extremal_points(
        cls,
        point1: volmdlr.Point3D,
        point2: volmdlr.Point3D,
        inner_radius: float,
        outer_radius: float,
        color: Tuple[float, float, float] = None,
        alpha: float = 1,
        name: str = "",
    ):
        """Deprecated class method. Use 'from_end_points' instead."""
        warnings.warn("Deprecated classmethod. Use 'from_end_points' instead.", DeprecationWarning)

        return cls.from_end_points(point1, point2, inner_radius, outer_radius, color, alpha, name)

    @classmethod
    def from_center_point_and_axis(
        cls,
        center_point: volmdlr.Point3D,
        axis: volmdlr.Vector3D,
        inner_radius: float,
        outer_radius: float,
        length: float,
        color: Tuple[float, float, float] = None,
        alpha: float = 1,
        name: str = "",
    ) -> 'HollowCylinder':
        """
        Create a hollow cylinder from a center point, an axis, radius, and length.

        :param center_point: The center point of the hollow cylinder
            (i.e. the middle point of the axis of the hollow cylinder).
        :type center_point: volmdlr.Point3D
        :param axis: The axis of revolution for the hollow cylinder.
        :type axis: volmdlr.Vector3D
        :param inner_radius: The inner radius of the hollow cylinder.
        :type inner_radius: float
        :param outer_radius: The outer radius of the hollow cylinder.
        :type outer_radius: float
        :param length: The length of the hollow cylinder.
        :type length: float
        :param color: The color of the hollow cylinder as an RGB tuple. Default is None.
        :type color: Tuple[float, float, float], optional
        :param alpha: The opacity of the hollow cylinder (0.0 to 1.0). Default is 1.0.
        :type alpha: float, optional
        :param name: The name of the hollow cylinder. Default is an empty string.
        :type name: str, optional

        :return: A HollowCylinder instance created from the specified center point, axis, radii, and length.
        :rtype: HollowCylinder
        """
        u_vector = axis.deterministic_unit_normal_vector()
        v_vector = axis.cross(u_vector)
        frame = volmdlr.Frame3D(center_point, u_vector, v_vector, axis)

        return cls(
            frame=frame,
            inner_radius=inner_radius,
            outer_radius=outer_radius,
            length=length,
            color=color,
            alpha=alpha,
            name=name,
        )

    def rotation(self, center: volmdlr.Point3D, axis: volmdlr.Vector3D, angle: float) -> 'HollowCylinder':
        """
        Hollow cylinder rotation.

        :param center: The rotation center.
        :type center: volmdlr.Point3D
        :param axis: The rotation axis.
        :type axis: volmdlr.Vector3D
        :param angle: The angle of rotation.
        :type angle: float

        :return: A new rotated HollowCylinder.
        :rtype: HollowCylinder
        """
        return self.__class__(
            frame=self.frame.rotation(center, axis, angle),
            length=self.length,
            inner_radius=self.inner_radius,
            outer_radius=self.outer_radius,
            color=self.color,
            alpha=self.alpha,
            name=self.name,
        )

    def translation(self, offset: volmdlr.Vector3D) -> 'HollowCylinder':
        """
        Hollow cylinder translation.

        :param offset: The translation vector.
        :type offset: volmdlr.Vector3D

        :return: A new translated HollowCylinder.
        :rtype: HollowCylinder
        """
        return self.__class__(
            frame=self.frame.translation(offset),
            length=self.length,
            inner_radius=self.inner_radius,
            outer_radius=self.outer_radius,
            color=self.color,
            alpha=self.alpha,
            name=self.name,
        )

    def frame_mapping(self, frame: volmdlr.Frame3D, side: str) -> 'HollowCylinder':
        """
        Changes frame_mapping and return a new HollowCylinder.

        side = 'old' or 'new'.
        """
        return HollowCylinder(
            frame=self.frame.frame_mapping(frame, side),
            inner_radius=self.inner_radius,
            outer_radius=self.outer_radius,
            length=self.length,
            color=self.color,
            alpha=self.alpha,
            name=self.name,
        )

    def copy(self, *args, **kwargs) -> 'HollowCylinder':
        """
        Creates a copy of HollowCylinder.

        :return: A copy of a current HollowCylinder.
        :rtype: HollowCylinder
        """
        return HollowCylinder(
            frame=self.frame.copy(),
            inner_radius=self.inner_radius,
            outer_radius=self.outer_radius,
            length=self.length,
            color=self.color,
            alpha=self.alpha,
            name=self.name,
        )


class Sweep(shells.ClosedShell3D):
    """
    Sweep a 2D contour along a Wire3D.

    """

    def __init__(self, contour2d: volmdlr.wires.Contour2D,
                 wire3d: volmdlr.wires.Wire3D, *,
                 color: Tuple[float, float, float] = None, alpha: float = 1,
                 name: str = ''):
        self.contour2d = contour2d
        self.wire3d = wire3d
        self.frames = []
        faces = self.shell_faces()
        shells.ClosedShell3D.__init__(self, faces, color=color,
                                      alpha=alpha, name=name)

    def to_dict(self, *args, **kwargs):
        """Custom serialization for performance."""
        dict_ = dc.DessiaObject.base_dict(self)
        dict_.update({'color': self.color,
                      'alpha': self.alpha,
                      'wire3d': self.wire3d.to_dict(),
                      'contour2d': self.contour2d.to_dict()
                      })

        return dict_

    def shell_faces(self):
        """
        Generates the shell faces.

        For now, it does not take into account rotation of sections.
        """
        def get_sweep_profile_section(profile_section_face, touching_faces):
            primitives_on_section_face = []
            for face in touching_faces:
                for prim in face.outer_contour3d.primitives:
                    if profile_section_face.edge3d_inside(prim):
                        primitives_on_section_face.append(prim)
            contour3d_ = volmdlr.wires.Contour3D.contours_from_edges(primitives_on_section_face)[0]
            return contour3d_

        # End  planar faces
        w = self.wire3d.primitives[0].unit_direction_vector(0.)
        u = self.wire3d.primitives[0].unit_normal_vector(0.)
        if not u:
            u = w.deterministic_unit_normal_vector()
        v = w.cross(u)

        start_plane = surfaces.Plane3D(
            volmdlr.Frame3D(self.wire3d.point_at_abscissa(0.), u, v, w))

        length_last_primitive = self.wire3d.primitives[-1].length()
        w = self.wire3d.primitives[-1].unit_direction_vector(length_last_primitive)
        u = self.wire3d.primitives[-1].unit_normal_vector(length_last_primitive)
        if not u:
            u = w.deterministic_unit_normal_vector()
        v = w.cross(u)

        end_plane = surfaces.Plane3D(
            volmdlr.Frame3D(self.wire3d.primitives[-1].point_at_abscissa(length_last_primitive), u, v, w))

        faces = [volmdlr.faces.PlaneFace3D(start_plane, surfaces.Surface2D(self.contour2d, []))]

        max_brectangle = max(self.contour2d.bounding_rectangle.bounds()) * 4
        cutting_face_contour = volmdlr.wires.Contour2D.from_points(
                    [volmdlr.Point2D(-max_brectangle, -max_brectangle),
                     volmdlr.Point2D(max_brectangle, -max_brectangle),
                     volmdlr.Point2D(max_brectangle, max_brectangle),
                     volmdlr.Point2D(-max_brectangle, max_brectangle)])
        new_faces = []
        last_end_tangent = self.wire3d.primitives[0].unit_direction_vector(0.)
        for i, wire_primitive in enumerate(self.wire3d.primitives):
            start_tangent = wire_primitive.unit_direction_vector(0.)
            normal = wire_primitive.unit_normal_vector(0.)
            if normal is None:
                normal = start_tangent.deterministic_unit_normal_vector()
            tangent_normal_orthonormal = start_tangent.cross(normal)

            if i == 0 or not start_tangent.is_close(last_end_tangent):
                new_faces = wire_primitive.sweep(self.contour2d, None)
            else:
                plane = surfaces.Plane3D(volmdlr.Frame3D(wire_primitive.start, normal,
                                                         tangent_normal_orthonormal, start_tangent))
                cutting_face = volmdlr.faces.PlaneFace3D(plane, surfaces.Surface2D(cutting_face_contour, []))
                section_contour = get_sweep_profile_section(cutting_face, new_faces)
                new_faces = wire_primitive.sweep(self.contour2d, section_contour)

            faces.extend(new_faces)
            last_end_tangent = wire_primitive.unit_direction_vector(wire_primitive.length())
        cutting_face = volmdlr.faces.PlaneFace3D(end_plane, surfaces.Surface2D(cutting_face_contour, []))
        contour3d = get_sweep_profile_section(cutting_face, new_faces)
        contour2d = cutting_face.surface3d.contour3d_to_2d(contour3d)
        end_face = volmdlr.faces.PlaneFace3D(end_plane, surfaces.Surface2D(contour2d, []))
        faces.append(end_face)
        return faces

    def frame_mapping(self, frame: volmdlr.Frame3D, side: str):
        """
        Changes frame_mapping and return a new Sweep.

        :param side: 'old' or 'new'
        """
        new_wire = self.wire3d.frame_mapping(frame, side)
        return Sweep(self.contour2d, new_wire, color=self.color,
                     alpha=self.alpha, name=self.name)

    def copy(self, deep=True, memo=None):
        """Creates a copy of the Sweep."""
        new_contour2d = self.contour2d.copy()
        new_wire3d = self.wire3d.copy()
        return Sweep(new_contour2d, new_wire3d, color=self.color,
                     alpha=self.alpha, name=self.name)


class Sphere(shells.ClosedShell3D):
    """
    Defines a sphere at a given position & radius.
    """

    def __init__(self, center: volmdlr.Point3D, radius: float,
                 color: Tuple[float, float, float] = None, alpha: float = 1.,
                 name: str = ''):
        self.center = center
        self.radius = radius
        self.position = center

        self.frame = volmdlr.Frame3D(center, volmdlr.X3D, volmdlr.Y3D, volmdlr.Z3D)
        spherical_surface = surfaces.SphericalSurface3D(self.frame, self.radius)
        spherical_face = volmdlr.faces.SphericalFace3D.from_surface_rectangular_cut(spherical_surface)
        shells.ClosedShell3D.__init__(self, faces=[spherical_face], color=color, alpha=alpha, name=name)

    def volume(self):
        """
        Computes the volume of the sphere.

        :return: sphere's volume (m³)
        """
        return 4 / 3 * math.pi * self.radius**3

    def point_belongs(self, point3d: volmdlr.Point3D, **kwargs) -> bool:
        """
        Returns if the point belongs to the sphere.

        :param point3d: volmdlr Point3D
        :return: True if the given point is inside the sphere, False otherwise
        """
        return self.center.point_distance(point3d) <= self.radius

    def frame_mapping(self, frame: volmdlr.Frame3D, side: str):
        """
        Changes frame_mapping and return a new Sphere.

        :param side: 'old' or 'new'
        """
        return Sphere(self.center.frame_mapping(frame, side), self.radius)

    def skin_points(self, resolution: float = 1e-3):
        """Gives points on the skin with respect to a resolution."""
        if resolution > 2 * self.radius:
            return []

        theta = 2 * math.asin(resolution / (2 * self.radius))

        nb_floor = int(math.pi / theta) + 1
        rota_theta = [n * theta for n in range(nb_floor)]

        point1 = self.center + volmdlr.X3D * self.radius

        skin_points = []

        for theta_ in rota_theta:
            pt_floor_init = point1.rotation(self.center, volmdlr.Y3D, theta_)

            if math.isclose(theta_, 0, abs_tol=1e-6) or math.isclose(theta_, math.pi, abs_tol=1e-6):
                skin_points.append(pt_floor_init)

            else:
                center_floor = volmdlr.Point3D(volmdlr.X3D.dot(pt_floor_init),
                                               self.center.y,
                                               self.center.z)

                r_floor = center_floor.point_distance(pt_floor_init)
                theta_floor = resolution / r_floor

                rota_theta_floor = [n * theta_floor for n in range(int(2 * math.pi / theta_floor) + 1)]

                if (2 * math.pi - rota_theta_floor[-1]) / theta_floor <= 0.1:
                    rota_theta_floor.pop()

                for tetha_f in rota_theta_floor:
                    pt_floor = pt_floor_init.rotation(center_floor, volmdlr.X3D, tetha_f)
                    skin_points.append(pt_floor)

        return skin_points

    def inner_points(self, resolution: float = 1e-3):
        """Gives points inside the sphere with a sub-sphere strategy."""
        in_points = [self.center]
        nb_spheres = int(self.radius / resolution)
        if nb_spheres == 0:
            return in_points

        spheres_radius = [n * resolution for n in range(1, nb_spheres + 1)]

        if (self.radius - spheres_radius[-1]) / resolution <= 0.1:
            spheres_radius.pop()

        for srad in spheres_radius:
            in_sphere = Sphere(self.center, srad)
            in_points.extend(in_sphere.skin_points(resolution=resolution))

        return in_points


class Measure3D:
    """
    Used to create a measure between two points in 3D.
    """

    def __init__(self, point1: volmdlr.Point3D, point2: volmdlr.Point3D,
                 color: Tuple[float, float, float] = (1., 0, 0)):
        self.point1, self.point2 = point1, point2
        self.color = color
        self.distance = (point1 - point2).norm()
        # self.bounding_box = self._bounding_box()

    # !!! no eq defined!
    def __hash__(self):
        return hash(self.point1) + hash(self.point2)


class BSplineExtrusion(volmdlr.core.Primitive3D):
    """
    Defines the extrusion of a BSpline.

    :param vectorextru: extrusion vector.
    """

    def __init__(self, obj, vectorextru: volmdlr.Vector3D, name: str = ''):
        self.obj = obj
        vectorextru = vectorextru.unit_vector()
        self.vectorextru = vectorextru
        if obj.__class__ is curves.Ellipse3D:
            self.points = obj.tessel_points
        else:
            self.points = obj.points

        volmdlr.core.Primitive3D.__init__(self, name=name)

    @classmethod
    def from_step(cls, arguments, object_dict, **kwargs):
        """
        Converts a step primitive to a BSplineExtrusion.

        :param arguments: The arguments of the step primitive. The last element represents the unit_conversion_factor.
        :type arguments: list
        :param object_dict: The dictionary containing all the step primitives
            that have already been instantiated.
        :type object_dict: dict
        :return: The corresponding BSplineExtrusion object.
        :rtype: :class:`volmdlr.primitives3d.BSplineExtrusion`
        """
        name = arguments[0][1:-1]
        if object_dict[arguments[1]].__class__ is curves.Ellipse3D:
            ell = object_dict[arguments[1]]
            vectextru = -object_dict[arguments[2]]
            return cls(ell, vectextru, name)

        if object_dict[arguments[1]].__class__ is volmdlr.edges.BSplineCurve3D:
            bsplinecurve = object_dict[arguments[1]]
            vectextru = object_dict[arguments[2]]
            return cls(bsplinecurve, vectextru, name)
        raise NotImplementedError  # to be adapted to bspline



# -*- coding: utf-8 -*-
"""
Setup install script for volmdlr

"""

import re
from os.path import dirname, isdir, join
from subprocess import CalledProcessError, check_output

import numpy as np
from setuptools import setup

from Cython.Build import cythonize  # isort: skip This prevent a build bug

tag_re = re.compile(r"\btag: %s([0-9][^,]*)\b")
version_re = re.compile("^Version: (.+)$", re.M)


def readme():
    with open("README.md") as f:
        return f.read()


def version_from_git_describe(version):
    if version[0] == "v":
        version = version[1:]

    # PEP 440 compatibility
    number_commits_ahead = 0
    if "-" in version:
        version, number_commits_ahead, commit_hash = version.split("-")
        number_commits_ahead = int(number_commits_ahead)

    # print('number_commits_ahead', number_commits_ahead)

    split_versions = version.split(".")
    if "post" in split_versions[-1]:
        suffix = split_versions[-1]
        split_versions = split_versions[:-1]
    else:
        suffix = None

    for pre_release_segment in ["a", "b", "rc"]:
        if pre_release_segment in split_versions[-1]:
            if number_commits_ahead > 0:
                split_versions[-1] = str(split_versions[-1].split(pre_release_segment)[0])
                if len(split_versions) == 2:
                    split_versions.append("0")
                if len(split_versions) == 1:
                    split_versions.extend(["0", "0"])

                split_versions[-1] = str(int(split_versions[-1]) + 1)
                future_version = ".".join(split_versions)
                return "{}.dev{}".format(future_version, number_commits_ahead)
            else:
                return ".".join(split_versions)

    if number_commits_ahead > 0:
        if len(split_versions) == 2:
            split_versions.append("0")
        if len(split_versions) == 1:
            split_versions.extend(["0", "0"])
        split_versions[-1] = str(int(split_versions[-1]) + 1)
        split_versions = ".".join(split_versions)
        return "{}.dev{}+{}".format(split_versions, number_commits_ahead, commit_hash)
    else:
        if suffix is not None:
            split_versions.append(suffix)

        return ".".join(split_versions)


# Just testing if get_version works well
assert version_from_git_describe("v0.1.7.post2") == "0.1.7.post2"
assert version_from_git_describe("v0.0.1-25-gaf0bf53") == "0.0.2.dev25+gaf0bf53"
assert version_from_git_describe("v0.1-15-zsdgaz") == "0.1.1.dev15+zsdgaz"
assert version_from_git_describe("v1") == "1"
assert version_from_git_describe("v1-3-aqsfjbo") == "1.0.1.dev3+aqsfjbo"


def get_version():
    # Return the version if it has been injected into the file by git-archive
    version = tag_re.search("$Format:%D$")
    if version:
        return version.group(1)

    d = dirname(__file__)

    if isdir(join(d, ".git")):
        cmd = "git describe --tags"
        try:
            version = check_output(cmd.split()).decode().strip()[:]

        except CalledProcessError:
            raise RuntimeError("Unable to get version number from git tags")

        return version_from_git_describe(version)
    else:
        # Extract the version from the PKG-INFO file.
        with open(join(d, "PKG-INFO")) as f:
            version = version_re.search(f.read()).group(1)

    # print('version', version)
    return version


setup(
    name="volmdlr",
    version=get_version(),
    description=" A volume modeler computation-oriented. Include rendering bindings.",
    long_description=readme(),
    long_description_content_type="text/markdown",
    keywords="volume, modeler, CAD",
    url="https://github.com/Dessia-tech/volmdlr",
    author="DessiA Technologies",
    author_email="root@dessia.tech",
    license="Creative Commons Attribution-Share Alike license",
    packages=[
        "volmdlr",
        "volmdlr.models",
        "volmdlr.utils",
        "volmdlr.nurbs"
    ],  # ,'volmdlr.primitives2D','volmdlr.primitives3D','volmdlr.geometry'],
    package_dir={},
    include_package_data=True,
    install_requires=[
        "packaging",
        "dessia_common>=0.10.0",
        "Cython>=3.0.0",
        "numpy<1.26.0", #  Wheels of 1.26.0 were not ready
        "matplotlib",
        "scipy<=1.11.1",
        "geomdl",
        "networkx",
        "triangle",
        "plot_data>=0.10.9",
        "kaitaistruct",
        "binaryornot",
        "sympy",
        "trimesh",
        "rtree",
        "gmsh",
    ],
    extras_require={"test": ["coverage"],
                    "doc": ["sphinx", "nbsphinx", "pydata_sphinx_theme", "nbformat", "nbconvert",
                            "sphinx_copybutton", "sphinx_design"]},
    classifiers=["Topic :: Scientific/Engineering",
                 "Topic :: Multimedia :: Graphics :: 3D Modeling",
                 "Development Status :: 5 - Production/Stable"],

    ext_modules=cythonize(["volmdlr/core_compiled.pyx",
                           "volmdlr/discrete_representation_compiled.py",
                           "volmdlr/nurbs/core.pyx",
                           "volmdlr/nurbs/helpers.pyx",
                           "volmdlr/nurbs/fitting.py"]),
    include_dirs=[np.get_include()],
    python_requires=">=3.9",
)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Module containing grid and relative objects.

"""

from typing import List

import numpy as npy

from dessia_common.core import DessiaObject  # isort: skip

import volmdlr
import volmdlr.wires


class Grid2D(DessiaObject):
    """
    A class defined with a list of points and characterized with a chosen direction.

    :param lists_points: A list of a list of points
    :type lists_points: List[List[volmdlr.Point2D]]
    :param direction: A direction
    :type direction: List[str]
    """

    def __init__(self, lists_points: List[List[volmdlr.Point2D]],
                 direction: List[str],
                 name: str = ''):

        self.lists_points = lists_points
        self.direction = direction
        DessiaObject.__init__(self, name=name)

    def displacement_compared_to(self, initial_grid2d):
        """
        Computes the deformation/displacement (dx,dy) of a grid2d based on another grid2d.

        :param initial_grid2d: A 2 dimensional grid
        :type initial_grid2d: :class:`volmdlr.grid.Grid2D`
        :return: The displacement of the 2 dimensional grid
        :rtype:
        """
        points_2d = initial_grid2d.points
        points_2d_deformed = self.points

        # Grid2D points displacement
        displacement = npy.ones(shape=(len(points_2d), 2))
        for i, displacement_i in enumerate(displacement):
            displacement_i[0] = points_2d_deformed[i][0] - points_2d[i][0]
            displacement_i[1] = points_2d_deformed[i][1] - points_2d[i][1]

        return displacement

    def find_direction_index(self, direction_axis: str):
        """
        Finds the index of a given direction_axis.

        :param direction_axis: 'x' or 'y'
        :type direction_axis: str
        :return: The direction index
        :rtype: int
        """

        try:
            index = self.direction.index('+' + direction_axis)
        except ValueError:
            index = self.direction.index('-' + direction_axis)

        return index

    @classmethod
    def from_points(cls, points, points_dim_1, direction, name: str = ''):
        """
        Defines a Grid2D given a list of points, number of points along the 1st dimension, and a direction.

        :param points:
        :type points: List[:class:`volmdlr.Point2D`]
        :param points_dim_1:
        :type points_dim_1: int
        :param direction:
        :type direction: List[str].
        :param name: object's name.
        :return:
        :rtype:
        """

        lists_points = [points[i:i + points_dim_1] for i in range(0, len(points), points_dim_1)]

        return cls(lists_points, direction, name=name)

    @classmethod
    def from_properties(cls, x_limits, y_limits, points_nbr,
                        direction=None, name: str = ''):
        """
        Defines Grid2d based on the given properties.

        :param x_limits: x_min and x_max
        :type x_limits: Tuple[float, float]
        :param y_limits: y_min and y_max
        :type y_limits: Tuple[float, float]
        :param points_nbr: Number of points along the x-axis and the y-axis
        :type points_nbr: Tuple[int, int]
        :param direction: Used for ordering the generated points
        :type direction: List[str].
        :param name: object's name.
        :return: The 2 dimensional grid
        :rtype: :class:`volmdlr.grid.Grid2D`
        """
        if direction is None:
            direction = ['+x', '+y']

        xmin, xmax = x_limits
        ymin, ymax = y_limits
        points_x, points_y = points_nbr

        directions_properties = {
            ('+x', '+y'): (xmin, xmax, ymin, ymax),
            ('-x', '+y'): (xmax, xmin, ymin, ymax),
            ('+y', '+x'): (xmin, xmax, ymin, ymax),
            ('-y', '+x'): (xmin, xmax, ymax, ymin),
            ('+x', '-y'): (xmin, xmax, ymax, ymin),
            ('-x', '-y'): (xmax, xmin, ymax, ymin),
            ('+y', '-x'): (xmax, xmin, ymin, ymax),
            ('-y', '-x'): (xmax, xmin, ymax, ymin)
        }

        grid2d = []
        points = []

        xmin, xmax, ymin, ymax = directions_properties[tuple(direction)]
        x = npy.linspace(xmin, xmax, points_x)
        y = npy.linspace(ymin, ymax, points_y)

        if direction in [['+x', '+y'], ['-x', '+y'], ['+x', '-y'], ['-x', '-y']]:
            for yi in y:
                for xi in x:
                    points.append(volmdlr.Point2D(xi, yi))

                grid2d.append(points)
                points = []

        elif direction in [['+y', '+x'], ['-y', '+x'], ['+y', '-x'], ['-y', '-x']]:
            for xi in x:
                for yi in y:
                    points.append(volmdlr.Point2D(xi, yi))

                grid2d.append(points)
                points = []

        return cls(lists_points=grid2d, direction=direction, name=name)

    def grid_pattern(self):
        """
        Defines a list of quadrilateral polygons defined based on Grid2d points.

        :return: The list of quadrilateral polygons
        :rtype: List[volmdlr.wires.ClosedPolygon2D]
        """

        quadrilateral_polygons = []
        length_1, length_2 = len(self.lists_points[0]), len(self.lists_points)
        for i in range(0, length_1 - 1):
            for j in range(0, length_2 - 1):
                quadrilateral_polygons.append(volmdlr.wires.ClosedPolygon2D(
                    [self.lists_points[i][j],
                     self.lists_points[i + 1][j],
                     self.lists_points[i + 1][j + 1],
                     self.lists_points[i][j + 1]]))

        return quadrilateral_polygons

    @property
    def limits_xy(self):
        """
        Finds the limits (min, max) of points along x & y direction_axis.

        :return:
        :rtype:
        """

        x_limits = (self.lists_points[0][0].x, self.lists_points[-1][-1].x)
        x_min, x_max = min(x_limits), max(x_limits)
        y_limits = (self.lists_points[0][0].y, self.lists_points[-1][-1].y)
        y_min, y_max = min(y_limits), max(y_limits)

        return ((x_min, x_max), (y_min, y_max))

    # @property
    # def x_limits(self):
    #     """
    #     find the limits (min, max) of points along x_direction_axis
    #     """

    #     x_limits = (self.lists_points[0][0].x, self.lists_points[-1][-1].x)
    #     x_min, x_max = min(x_limits), max(x_limits)

    #     return (x_min, x_max)

    # @property
    # def y_limits(self):
    #     """
    #     find the limits (min, max) of points along y_direction_axis
    #     """

    #     y_limits = (self.lists_points[0][0].y, self.lists_points[-1][-1].y)
    #     y_min, y_max = min(y_limits), max(y_limits)

    #     return (y_min, y_max)

    @property
    def points(self):
        """
        Returns all the points in lists_points in just one list.

        :return: The flattened list of points
        :rtype: List[:class:`volmdlr.Point2D`]
        """

        points = []
        for list_point in self.lists_points:
            points.extend(list_point)
        return points

    @property
    def points_xy(self):
        """
        Finds how many points there are along x & y direction_axis.

        :return: Two counts, one for the x direction_axis and another one for
            the y direction_axis
        :rtype: Tuple[int, int]
        """

        index = self.find_direction_index(direction_axis='x')
        if index == 0:
            points_x = len(self.lists_points[0])
            points_y = len(self.lists_points)
        else:
            points_x = len(self.lists_points)
            points_y = len(self.lists_points[0])

        return (points_x, points_y)

    # @property
    # def points_x(self):
    #     """
    #     find how many points there are along x_direction_axis
    #     """

    #     index = self.find_direction_index(direction_axis = 'x')
    #     if index == 0:
    #         points_x = len(self.lists_points[0])
    #     else:
    #         points_x = len(self.lists_points)

    #     return points_x

    # @property
    # def points_y(self):
    #     """
    #     find how many points there are along y_direction_axis
    #     """

    #     index = self.find_direction_index(direction_axis = 'y')
    #     if index == 0:
    #         points_y = len(self.lists_points[0])
    #     else:
    #         points_y = len(self.lists_points)

    #     return points_y



import unittest

import numpy as npy

import volmdlr as vm
from volmdlr import cloud, faces, shells


class TestCloud(unittest.TestCase):
    points_cloud = cloud.PointCloud3D([vm.Point3D(0, 0, 0), vm.Point3D(10, 10, 10),vm.Point3D(-2.5, 1.2, 0)])

    point1 = vm.Point3D(1, 1, 1)
    point2 = vm.Point3D(1, 2, 1)
    point3 = vm.Point3D(1, 1, 2)
    point4 = vm.Point3D(2, 1, 1)
    point5 = vm.Point3D(2, 2, 1)
    point6 = vm.Point3D(1, 2, 2)
    point7 = vm.Point3D(2, 1, 2)
    point8 = vm.Point3D(2, 2, 2)
    open_triangle_shell = shells.OpenTriangleShell3D([faces.Triangle3D(point1, point2, point3),
                                                     faces.Triangle3D(point3, point4, point5),
                                                     faces.Triangle3D(point5, point6, point7),
                                                     faces.Triangle3D(point7, point8, point6)])

    def test_shell_distances(self):
        images, distances, nearest_shells = self.points_cloud.shell_distances(self.open_triangle_shell)

        self.assertIsInstance(images, cloud.PointCloud3D)
        self.assertIsInstance(distances, list)
        self.assertIsInstance(nearest_shells, list)

        self.assertEqual(nearest_shells, [0, 3, 0])
        self.assertEqual(round(distances[0], 2), 1.73)
        self.assertEqual(round(distances[1], 2), 13.86)
        self.assertEqual(round(distances[2], 2), 3.64)

    def test_shell_distances_ndarray(self):
        images, distances, nearest_shells = self.points_cloud.shell_distances_ndarray(self.open_triangle_shell)

        self.assertIsInstance(images, npy.ndarray)
        self.assertIsInstance(distances, npy.ndarray)
        self.assertIsInstance(nearest_shells, npy.ndarray)

        self.assertTrue(npy.all(nearest_shells == npy.array([0, 3, 0])))
        self.assertEqual(round(distances[0], 2), 1.73)
        self.assertEqual(round(distances[1], 2), 13.86)
        self.assertEqual(round(distances[2], 2), 3.64)

if __name__ == '__main__':
    unittest.main()
